#
#  SCCaller: A framework for accessing Oracle Storage Connect
#
#  Copyright (C) 2010 Oracle.  All rights reserved.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#

"""A high-level interface to Oracle Storage Connect"""

import string
import os
import os.path
import tempfile
import time
from bsddb.db import *


__all__ = ['SCCallerCache']


global_db = None


class SCCallerCacheDB(object):

    """A convenience wrapper around the Berekely DB

    This is intended for the internal use of SCCallerCache.

    """

    def __init__(self, home=None):
        self.home = home
        if not self.home:
            self.home = tempfile.mkdtemp(prefix='sccaller')
        if not os.path.exists(self.home):
            os.mkdir(self.home)
        self._env = DBEnv()
        self._env.open(self.home, DB_INIT_CDB|DB_INIT_MPOOL|DB_CREATE, 0600)

    def store(self, database, key, data):
        db = DB(self._env)
        try:
            db.open('sccaller', database, DB_BTREE, DB_CREATE, 0600)
        except DBError, e:
            # Need to handle recovery.
            # Log the error, but otherwise failing to store is just slower
            return
        db.put(key, data)
        db.close()

    def retrieve(self, database, key):
        db = DB(self._env)
        try:
            db.open('sccaller', database, DB_BTREE, DB_RDONLY, 0600)
        except DBError, e:
            # Need to handle recovery.
            # Log the error, but otherwise failing to get is like it doesn't
            # exist
            return None
        data = db.get(key)
        db.close()
        return data

    def delete(self, database, key):
        db = DB(self._env)
        try:
            db.open('sccaller', database, DB_BTREE, DB_CREATE, 0600)
        except DBError, e:
            # Need to handle recovery.
            # Log the error, but otherwise failing to get is like it doesn't
            # exist
            return
        db.delete(key)
        db.close()

    def clear(self, database):
        db = DB(self._env)
        try:
            db.open('sccaller', database, DB_BTREE, DB_CREATE, 0600)
        except DBError, e:
            # Need to handle recovery.
            # Log the error, but otherwise failing to get is like it doesn't
            # exist
            return None
        for key in db.keys():
            db.delete(key)
        db.close()


class SCCallerCache(object):

    """A host-persistent cache for Storage Connect plug-ins

    This cache stores its values on disk, allowing for multiple processes
    to share the cache.  See OSCCache for details of the API.

    """

    def __init__(self, home=None):
        """Create a new cache object"""
        global global_db
        if global_db:
            if not home or (global_db.home == home):
                self._db = global_db
            else:
                self._db = SCCallerCacheDB(home)
        else:
            self._db = SCCallerCacheDB(home)
            global_db = self._db
        self.home = self._db.home

    def _check_expire(self, ss_uuid, key):
        try:
            expire_data = self._db.retrieve('%s.expire' % ss_uuid, key)
        except DBNotFoundError, e:
            return
        if not expire_data:
            return
        expire_time = string.atol(expire_data)
        if time.time() > expire_time:
            self._db.delete('%s.data' % ss_uuid, key)
            self._db.delete('%s.expire' % ss_uuid, key)

    def set(self, ss_uuid, key, value, expire_seconds=30):
        if value != None:
            # Round to next second
            expire_time = long(time.time() + expire_seconds + 0.9)
            self._db.store('%s.data' % ss_uuid, key, value)
            self._db.store('%s.expire' % ss_uuid, key, str(expire_time))
        else:
            self._db.delete('%s.data' % ss_uuid, key)
            self._db.delete('%s.expire' % ss_uuid, key)

    def get(self, ss_uuid, key):
        self._check_expire(ss_uuid, key)
        try:
            data = self._db.retrieve('%s.data' % ss_uuid, key)
        except DBNotFoundError, e:
            data = None
        return data

    def extend(self, ss_uuid, key, extend_seconds):
        self._check_expire(ss_uuid, key)
        try:
            expire_data = self._db.retrieve('%s.expire' % ss_uuid, key)
        except DBNotFoundError, e:
            return
        if not expire_data:
            return
        expire_time = long(string.atol(expire_data) + extend_seconds + 0.9)
        self._db.store('%s.expire' % ss_uuid, key, str(expire_time))

    def clear(self, ss_uuid):
        self._db.clear('%s.data' % ss_uuid)
        self._db.clear('%s.expire' % ss_uuid)
