#
#  SCCaller: A framework for accessing Oracle Storage Connect
#
#  Copyright (C) 2010 Oracle.  All rights reserved.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#

"""A high-level interface to Oracle Storage Connect"""

import sys
import copy
import os.path
import asyncore
import subprocess
import logging
from errno import *
try:
    import json
except ImportError:
    import simplejson as json
try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

from sccaller.chat import StdinoutChat



__all__ = ['SCCaller']

worker_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                           'worker.py'))
worker_cmd = [sys.executable, worker_path]

SCRUB_LIST = ['password', 'passwd']


def scrub_dict(d):
    for key in d:
        for word in SCRUB_LIST:
            if word in key:
                d[key] = '******'


def scrub_and_log(logger, result, err):
    result_copy = copy.deepcopy(result)
    if type(result_copy) == dict:
        scrub_dict(result_copy)
    logger.debug('Result [%s, %s]' % (result_copy, err))


class SCProcess(StdinoutChat):
    def __init__(self, map=None, logging_level=0, logging_name=None, cache_notifier=None ):
        global worker_cmd
        self.callback = None
        self.cache_notifier = cache_notifier
        self._set_called = {}
        self.ibuffer = []
        self.packet_string = ''
        self.packet = None
        self.set_logging(logging_level, logging_name)
        self.set_terminator('\r\n\r\n')
        self.worker = subprocess.Popen(worker_cmd,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       close_fds=True)
        self.pid = self.worker.pid
        StdinoutChat.__init__(self, stdin_fd=self.worker.stdout,
                              stdout_fd=self.worker.stdin, map=map)
        # StdinoutChat will dup() the descriptors.  Close the
        # originals.
        self.worker.stdout.close()
        self.worker.stdin.close()

    def set_logging(self, level=0, name=None):
        # self.logger is for SCProcess messages
        self._logging_level = level
        self.logger = logging.getLogger('sccaller.caller.SCProcess')
        self.logger.setLevel(level)
        self.logger.addHandler(NullHandler())
        if not name:
            name = 'sccaller.caller.SCProcess(<No name>)'
        # self.plugin_logger is for messages received from the worker
        self.plugin_logger = logging.getLogger(name)
        self.plugin_logger.setLevel(self._logging_level)
        self.plugin_logger.addHandler(NullHandler())

    def log(self, message):
        self.logger.info(message)

    def log_info(self, message, type='info'):
        level = logging.getLevelName(type.upper())
        if isinstance(level, int):
            self.logger.log(level, message)

    def collect_incoming_data(self, data):
        self.ibuffer.append(data)

    def found_terminator(self):
        self.packet_string = ''.join(self.ibuffer)
        self.ibuffer = []
        self.packet = json.loads(self.packet_string)
        self.logger.log(logging.DEBUG / 2,
                        'Got packet "%s"\n' % self.packet_string)
        self.handle_packet()

    def push_call(self, operation, args, callback):
        if self.callback:
            raise ValueError('Operation already in progress')
        if operation.startswith('set-'):
            # If the caller is sending more than one operation to this
            # worker, we only want to call each set-* API once.  Keep
            # track so that the caller doesn't have to.
            # XXX Is there a way to compare the original with the later
            #     set- call?  I think so for the set-plugin name, but
            #     not for the set-server record.
            set_key = operation[len('set-'):]
            if self._set_called.has_key(set_key):
                callback(None, None)
                return
            else:
                self._set_called[set_key] = True
        self.callback = callback
        if not args:
            args = []
        call_packet = {
            'magic': 'SCCALLER',
            'major': 1,
            'operation': operation,
            'arguments': args,
        }
        self.push('%s\r\n\r\n' % json.dumps(call_packet,
                                            separators=(',',':')))

    def handle_packet(self):
        try:
            if self.packet['magic'] != 'SCCALLER':
                raise ValueError('Invalid packet: %s' % self.packet_string)
            if self.packet['major'] != 1:
                raise ValueError('Unknown packet version: %d' %
                                 self.packet['major'])
            if self.packet.has_key('log_message'):
                msg = self.packet['log_message']
                level = int(self.packet['log_level'])
                self.plugin_logger.log(level, msg)
            elif self.packet.has_key('result'):
                result = self.packet['result']
                e = self.packet['error']
                self.handle_result(result, e)
            elif self.packet.has_key('cache'):
                if self.cache_notifier:
                    self.logger.debug('Cache %s' % self.packet['cache'])
                    self.cache_notifier(self.packet['cache'])        
            else:
                raise ValueError('Unknown packet type: %s' %
                                 self.packet_string)
        except Exception, e:
            self.handle_result(None, e)
        self.packet_string = ''
        self.packet = None

    def handle_result(self, result, e):
        scrub_and_log(self.logger, result, e)
        callback = self.callback
        self.callback = None
        callback(result, e)


class SCCaller(object):

    """A high-level interface to Oracle Storage Connect

    SCCaller provides synchronous and asynchronous access to the
    storage operations provided by Oracle Storage Connect (OSC) in a
    pythonic fashion.  It aims for isolation from the specific
    ABI of the various storage plug-ins, instead providing a consistent
    API for users.

    An SCCaller instance is specific to a given OSC plug-in.  The
    instance can only perform one operation at a time; more than one
    SCCaller instance can be created for each plug-in.  Like all OSC
    interfaces, SCCaller does not keep track of storage objects from
    call to call.  All SCCaller operations are preformed by a
    sub-process; this keeps the various plug-ins out of the calling
    program's address space.

    Synchronous operation is the simplest.  Given a server record
    (as documented by OSC) and the name of the storage plug-in to use,
    one can easily retrieve the server's status and the list of LUNs
    on the server:

    >>> caller = SCCaller(plugin_name)
    >>> status = caller.get_server_status(server)
    >>> luns = caller.list_luns(server)

    Advanced features of the interface are accessed via keyword
    arguments available to all methods.

    Asynchronous operation is accessed via the callback keyword
    argument.  This specifies a callback function of the form
    callback(result, e).  The result argument contains the value that
    would be returned by the synchronous operation.  The e argument
    contians an exception if one occurred.  If e is not None, result
    will be None.

    If a callback is provided, the method will not block on the
    operation.  Instead, it will return a tuple of (read_fd, write_fd).
    This is the communication channel to the sub-process performing the
    operation.  The calling program must integrate these descriptors
    into its own event loop.  When either of the descriptors is ready,
    the program calls the dispatch() method.  SCCaller handles the
    rest.  When the operation completes, the callback will be fired:

    >>> def callback(result, e):
    >>>    global done
    >>>    if e:
    >>>        print 'Operation failed: %s' % str(e)
    >>>    else:
    >>>        print 'Status is: %s' % str(result)
    >>>    done = True

    >>> caller = SCCaller(plugin_name)
    >>> done = False
    >>> fds = caller.get_server_status(server, callback)
    >>> while not done:
    >>>    r = [fds[0]]
    >>>    w = [fds[1]]
    >>>    e = fds[:]
    >>>     try:
    >>>         r, w, e = select.select(r, w, e, 30.0)
    >>>     except select.error, err:
    >>>         if err.args[0] != EINTR:
    >>>             print 'Select failed: %s' % str(err)
    >>>             break
    >>>         else:
    >>>             continue
    >>>     caller.dispatch()

    The trivial select() loop above would be replaced by the calling
    program's event loop.

    For convenience, the dispatch method also returns True if the
    operation is still ongoing and False if it has completed.  Programs
    may make use of this in their select loops rather than passing
    variables back from their callbacks.

    """

    def __init__(self, plugin_name, logging_level=0, cache_notifier=None):
        """Create a new caller to access a Storage Connect plug-in

           @param plugin_name: the plugin this caller will use
           @param logging_level: the logging level for this caller
               (see set_logging_level())

        """
        self.plugin_name = plugin_name
        self.worker = None
        self._map = None
        self._multi_op = False
        self._in_loop = False
        self.callback = None
        self.result = None
        self.cache_notifier = cache_notifier
        self.op_list = None
        self.set_logging_level(logging_level)

    def set_logging_level(self, level=0):
        """Set the logging level for this caller

        Normally a library module would simply log everything at the
        highest level it cares about, letting the user filter out
        unneeded messages.  However, we want to avoid extra traffic
        from the worker process if we can help it, so SCCaller takes
        an explicit logging level.  It will pass this level down to
        the worker.

        @param logging_level: A logging level as provided to
            Logger.setLevel()

        """
        self._logging_level = level
        # self.logger is for SCCaller messages
        self.logger = logging.getLogger('sccaller.caller.SCCaller')
        self.logger.setLevel(self._logging_level)
        self.logger.addHandler(NullHandler())

    def _error_to_exception(self, e):
        """Convert Python exceptions for transmission.

        We don't pass python errors across the communication channel.
        Instead, we pass (errno, message) tuples.  Turn them back into
        exceptions.

        """

        import OSCPlugin
        errno, message = e
        if errno == EINVAL:
            return ValueError(message)
        elif errno == ENXIO:
            return TypeError(message)
        elif errno == EOPNOTSUPP:
            if not message:
                message = 'Operation not implemented'
            return NotImplementedError(message)
        elif errno == ESRCH:
            return ValueError(message)
        elif errno == ENODATA:
	    parts = message.split(":", 1)
            if hasattr(OSCPlugin, parts[0]):
                excpt = getattr(OSCPlugin, parts[0])
                return excpt(parts[1])
            else:
                return ValueError(message)
        else:
            return OSError(errno, message)

    def _set_ops(self, *op_list):
        """Set the operation list for the next call

        Fills in the operations list, which is a list of [op, args] pairs.
        Cannot be set if an operation is in progress.

        @param op_list: List of [op, args] pairs

        """
        if self.op_list != None:
            raise ValueError('Operation already in progress')
        self.op_list = list(op_list)
        if self._logging_level:
            self.op_list.insert(0, ['set-logger', [self._logging_level,
                                                   self.plugin_name]])

    def _start_worker(self):
        if not self.worker:
            self._map = {}
            name = self.plugin_name
            if not name:
                name = '%s.%s(<No plugin>)' % (self.__class__.__module__,
                                               self.__class__.__name__)
            self.worker = SCProcess(map=self._map,
                                    logging_level=self._logging_level,
                                    logging_name=name,
                                    cache_notifier=self.cache_notifier)
            self.logger.debug('Started worker process %d' % self.worker.pid)
        elif not self._multi_op:
            raise ValueError('Operation already in progress')

    def _stop_worker(self):
        if self.worker:
            self._map = None
            pid = self.worker.pid
            self.worker.close()
            self.worker = None
            self.logger.debug('Stopped worker process %d' % pid)

    def _multi_op_start(self):
        """Run multiple operations through one worker

        Some subclasses need to run multiple operations through one
        worker process.  Flag this case.  They must finish with
        _multi_op_end()

        """
        self._multi_op = True

    def _multi_op_end(self):
        """Finish multiple operations

        Some subclasses need to run multiple operations through one
        worker process.  This turns off the flag or closes the worker
        if there is no work pending.

        """
        if not self._multi_op or self._in_loop:
            return
        self._multi_op = False
        if self.worker and (self.op_list == None):
            self._stop_worker()

    def _op_callback(self):
        """Clear our state and send the result to the original caller."""
        callback = self.callback
        result, e = self.result
        self.callback = None
        self.result = None
        self.op_list = None
        if e:
            e = self._error_to_exception(e)
        callback(result, e)

    def _call_callback(self, result, e):
        self.result = (result, e)
        if e or not self.op_list:
            self._op_callback()
        else:
            (operation, arguments) = self.op_list.pop(0)
            self.worker.push_call(operation, arguments, self._call_callback)

    def _sync_op_callback(self, data, result, e):
        data.extend([result, e])
        if not self._multi_op:
            self._stop_worker()

    def _run_sync_ops(self):
        self._start_worker()
        data = []
        self.callback = lambda result, e: self._sync_op_callback(data,
                                                                 result, e)
        (operation, arguments) = self.op_list.pop(0)
        self.worker.push_call(operation, arguments, self._call_callback)
        while not len(data):
            asyncore.loop(map=self._map, count=1)
        result, e = data
        if e:
            raise e
        return result

    def dispatch(self):
        """Perform pending communication

        Read or write to the sub-process when a communication channel
        is ready.  If the operation completes, the callback will be
        called.

        @returns: True if the operation is still ongoing, False if it is
            complete

        """
        self._in_loop = True
        asyncore.loop(map=self._map, count=1)
        self._in_loop = False
        if (self.op_list == None) and not self._multi_op:
            self._stop_worker()
            return False
        return True

    def _run_async_ops(self):
        self._start_worker()
        (operation, arguments) = self.op_list.pop(0)
        self.worker.push_call(operation, arguments, self._call_callback)
        return self.worker.get_fds()

    def _run_ops(self, **kwargs):
        if kwargs.has_key('callback'):
            self.callback = kwargs['callback']
            return self._run_async_ops()
        else:
            return self._run_sync_ops()

    def list_plugins(self, **kwargs):
        """List the plug-ins that are installed

        @returns: a list of plug-in names that can be passed to SCCaller
            instances.

        """
        self._set_ops(['list-plugins', None])
        return self._run_ops(**kwargs)

    def get_plugin_attributes(self, *attributes, **kwargs):
        """Get one or more attributes of the plug-in

        @param attributes: one or more attributes to query.  Shell
            wildcards are supported.  "*" selects all attributes.
        @returns: a dict of the attributes and their values

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['get-plugin-attributes', attributes],
        )
        return self._run_ops(**kwargs)

    def validate(self, server, **kwargs):
        """Validate a server record

        Quickly check that the server record has all the information
        needed to contact the server.  This does not check that the
        information is valid, just that all necessary fields are present.

        @param server: The server record to validate

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['validate-server', [server]],
        )
        return self._run_ops(**kwargs)

    def get_server_info(self, server, **kwargs):
        """Get information about the storage

        Query the server for status information.

        @param server: the record describing the server to query
        @returns: a populated server record

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-server-info', None],
        )
        return self._run_ops(**kwargs)

    def get_server_attributes(self, server, *attributes, **kwargs):
        """Get one or more attributes of the server

        Like get_plugin_attributes(), except it gets server attributes.
        Servers supply the same capability.* attributes, but they may not
        provide every capability the plug-in can support.  This may be due
        to limited hardware or restricted licensing.

        @param server: the record describing the server to query
        @param attributes: one or more attributes to query.  Shell
            wildcards are supported.  "*" selects all attributes.
        @returns: a dict of the attributes and their values

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-server-attributes', attributes],
        )
        return self._run_ops(**kwargs)

    def get_server_status(self, server, **kwargs):
        """Quickly query a server's status

        This is a quick operation to retreive a server's status.  It
        may be able to short-circuit some of te processing done by
        get_server_info().

        @param server: the record describing the server to query
        @returns: the server's status

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-server-status', None],
        )
        return self._run_ops(**kwargs)

    def get_storage_names(self, server, **kwargs):
        """Get storage names associated with this server

        Some storage is controled by a concentrator machine.  This is
        a machine that controls multiple storage servers.  All operations
        are directed to the management machine.  The storage name tells
        the management machine which storage server to operate on.

        Whether a storage type needs to specify a storage name can be
        queried via the "capability.storage-name-required" attribute.

        @param server: the record describing the server to query
        @returns: the list of storage names the management machine knows.
            If a storage server is being accessed directly, this will
            return the empty list.

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-storage-names', None],
        )
        return self._run_ops(**kwargs)

    def list_luns(self, server, **kwargs):
        """Retrieve a list of all whole LUNs on this server

        @param server: the record describing the server to query
        @returns: the list of whole LUNs

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-luns', None],
        )
        return self._run_ops(**kwargs)

    def list_clones(self, server, *parents, **kwargs):
        """Retrieve a list of clones from the server

        @param server: the record describing the server to query
        @param parents: if present, only return clones of the given
            parent(s)
        @returns: the list of matching clones

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-clones', parents],
        )
        return self._run_ops(**kwargs)

    def list_snapshots(self, server, *parents, **kwargs):
        """Retrieve a list of snapshots from the server

        @param server: the record describing the server to query
        @param parents: if present, only return snapshots of the given
            parent(s)
        @returns: the list of matching snapshots

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-snapshots', parents],
        )
        return self._run_ops(**kwargs)

    def get_element_status(self, server, element, **kwargs):
        """Quickly query an element's status

        This is a quick operation to retreive a element's status.  It
        may be able to short-circuit some of te processing done by
        get_element_info().

        @param server: the record describing the server to query
        @param element: the record describing the element to query
        @returns: the element's status

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-element-status', [element]],
        )
        return self._run_ops(**kwargs)

    def get_element_info(self, server, element, **kwargs):
        """Get information about a storage element

        Query the server for information about a storage element.

        @param server: the record describing the server to query
        @param element: the record describing the element to query
        @returns: an updated storage element record

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-element-info', [element]],
        )
        return self._run_ops(**kwargs)

    def get_element_attributes(self, server, element, *attributes,
                               **kwargs):
        """Get one or more attributes of a storage element

        Like get_server_attributes(), except it gets the attributes of a
        storage element.  Elements have a small number of attributes,
        mostly a subset of the capability.* attributes

        @param server: the record describing the server to query
        @param element: the record describing the element to query
        @param attributes: one or more attributes to query.  Shell
            wildcards are supported.  "*" selects all attributes.
        @returns: a dict of the attributes and their values

        """
        args = [element]
        args.extend(attributes)
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-element-attributes', args],
        )
        return self._run_ops(**kwargs)

    def list_access_groups(self, server, *elements, **kwargs):
        """Retreive the list of access groups on the server

        @param server: the record describing the server to query
        @param elements: if present, only return the access groups of
            the given elements(s)
        @returns: the list of access groups

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-access-groups', elements],
        )
        return self._run_ops(**kwargs)

    def create_access_group(self, server, name, *entries, **kwargs):
        """Create a new access group on the storage server

        Creates the access group name with the optionally specified entries.
        Passing an already-eisting group is an error.

        @param server: the record describing the server to access
        @param name: the name of the new access group
        @param entries: if present, access entries that should be added
            to the new group

        """
        args = [name]
        args.extend(entries)
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['create-access-group', args],
        )
        return self._run_ops(**kwargs)

    def remove_access_group(self, server, name, **kwargs):
        """Remove an access group from the storage server

        Removes the access group with the specified name.

        @param server: the record describing the server to access
        @param name: the name of the access group to remove

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['remove-access-group', [name]],
        )
        return self._run_ops(**kwargs)

    def rename_access_group(self, server, name, new_name, **kwargs):
        """Rename an access group on the storage server

        @param server: the record describing the server to access
        @param name: the name of the access group to rename
        @param new_name: the new name for the access group

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['rename-access-group', [name, new_name]],
        )
        return self._run_ops(**kwargs)

    def add_hosts_to_access_group(self, server, name, *entries, **kwargs):
        """Add initiator identifiers to an access group

        @param server: the record describing the server to access
        @param name: the access group these entries will be added to
        @param entries: new initiator identifiers for the access group

        """
        args = [name]
        args.extend(entries)
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['add-hosts-to-access-group', args],
        )
        return self._run_ops(**kwargs)

    def remove_hosts_from_access_group(self, server, name, *entries,
                                       **kwargs):
        """Remove initiator identifiers from an access group

        @param server: the record describing the server to access
        @param name: the access group to remove entries from
        @param entries: initiator identifiers to remove

        """
        args = [name]
        args.extend(entries)
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['remove-hosts-from-access-group', args],
        )
        return self._run_ops(**kwargs)

    def add_elements_to_access_group(self, server, name, *elements,
                                     **kwargs):
        """Add storage elements to an access group

        @param server: the record describing the server to access
        @param name: the access group these elements will be added to
        @param elements: new storage elements for the access group

        """
        args = [name]
        args.extend(elements)
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['add-elements-to-access-group', args],
        )
        return self._run_ops(**kwargs)

    def remove_elements_from_access_group(self, server, name, *elements,
                                          **kwargs):
        """Remove storage elements from an access group

        @param server: the record describing the server to access
        @param name: the access group to remove elements from
        @param elements: the storage elements to remove

        """
        args = [name]
        args.extend(elements)
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['remove-elements-from-access-group', args],
        )
        return self._run_ops(**kwargs)

    def iscsi_sendtargets(self, server, **kwargs):
        """Query targets for an iSCSI server

        Discovers the targets available on the specified iSCSI server.

        @param server: the record describing the server to access
        @returns: the list of available targets

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['iscsi-sendtargets', None],
        )
        return self._run_ops(**kwargs)

    def iscsi_start(self, server, *targets, **kwargs):
        """Start iSCSI by logging into the given targets

        Logs into the server with the given list of iSCSI targets.
        These are usually discovered via iscsi_sendtargets().

        @param server: the server to log in to
        @param targets: the targets to log in to

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['iscsi-start', targets],
        )
        return self._run_ops(**kwargs)

    def iscsi_stop(self, server, *targets, **kwargs):
        """Stop iSCSI by logging out of the given targets

        Logs out of the server with the given list of iSCSI targets.
        These are usually discovered via iscsi_sendtargets().

        @param server: the iSCSI server to log out of
        @param targets: the targets to log out of

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['iscsi-stop', targets],
        )
        return self._run_ops(**kwargs)

    def lun_scan(self, server, *targets, **kwargs):
        """Rescan the local bus for LUNs

        Scan the "local" bus for LUNs provided by this server.  If the
        server requires specific targets (eg, iSCSI target names
        discovered by iscsi_sendtargets()), specify them as arguments.

        Note that this call makes the OS perform the scan.  It does not
        return information about the resulting devices.  For that, use
        XXXX.

        @param server: the server to scan for attached devices
        @param targets: the targets for scanning if required

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['lun-scan', targets],
        )
        return self._run_ops(**kwargs)

    def map_devices(self, server, element, **kwargs):
        """Scan the local system's device tree for this element

        Scan the "local" system device tree for all devices that refer
        to this storage element.

        @param server: the server the element belongs to
        @param element: the storage element in question
        @returns: a list of local device paths

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['map-devices', [element]],
        )
        return self._run_ops(**kwargs)

    def element_online(self, server, element, **kwargs):
        """Bring the specified storage element online

        Some storage servers support online and offline states for
        their storage elements.  Before an element can be accessed,
        it must be in the online state.

        @param server: the server containing the storage element
        @param element: the storage element to bring online

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['element-online', [element]],
        )
        return self._run_ops(**kwargs)

    def element_offline(self, server, element, **kwargs):
        """Bring the specified storage element offline

        Some storage servers support online and offline states for
        their storage elements.  Often before an element can be removed
        or modified, it must be in the online state.

        @param server: the server containing the storage element
        @param element: the storage element to bring offline

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['element-offline', [element]],
        )
        return self._run_ops(**kwargs)

    def get_qos_list(self, server, element=None, **kwargs):
        """Retrieve the Quality of Service levels known to the server

        Some storage can control the behavior of its elements via
        Quality of Service (QoS) designations.  Some storage elements
        (such as filesystems) can even have different QoS designations.

        @param server: the server to query
        @param element: the element to query
        @returns: a list of dicts describing the QoS levels

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-qos-list', [element]],
        )
        return self._run_ops(**kwargs)

    def set_element_qos(self, server, element, qos_value, **kwargs):
        """Set the Quality of Service level on a storage element

        If the array supports dynamic configuration of Quality of
        Service (QoS), this call sets the QoS of the specified element.

        @param server: the server to act on
        @param element: the storage element to modify
        @param qos_value: the QoS value to set on the storage element.
            This is the "value" field of a dict returned by
            get_qos_list().

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['set-element-qos', [element, qos_value]],
        )
        return self._run_ops(**kwargs)

    def create_snapshot(self, server, element, snapshot_name=None,
                        **kwargs):
        """Create a snapshot of the given storage element

        Storage Connect defines a "snapshot" as a read-only thin copy
        of an existing storage element.  Snapshots may not be
        presentable to hosts.

        @param server: the server to act on
        @param element: the storage element to snapshot
        @param snapshot_name: a name for the new snapshot.  If not
            provided, a name will be generated.
        @returns: the new snapshot element

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['create-snapshot', [element, snapshot_name]],
        )
        return self._run_ops(**kwargs)

    def create_multi_element_snapshot(self, server, elements,
                                      snapshot_name=None, **kwargs):
        """Create a snapshot of the storage elements listed

        Some storage servers can snapshot more than one storage
        element into a single snapshot.  This is treated as a single
        entity for backup and restore.

        @param server: the server to act on
        @param element: the storage elements to snapshot
        @param snapshot_name: a name for the new snapshot.  If not
            provided, a name will be generated.
        @returns: the new snapshot element

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['create-multi-element-snapshot', [elements, snapshot_name]],
        )
        return self._run_ops(**kwargs)

    def create_clone(self, server, element, clone_name=None,
                     qos_value=None, snapshot_element=None,
                     plugin_options=None, **kwargs):
        """Create a clone of the given storage element

        Storage Connect defines a "clone" as a writable thin copy
        of an existing storage element.  Clones are treatable like LUNs,
        they can be presented to hosts.  Some storage allows cloning of
        LUNs directly.  Other storage only allows clones of snapshots.
        To clone a snapshot, provide the original LUN via element and the
        snapshot in snapshot_element.

        Storage that allows snapshots and only clons of snapshots can
        clone a LUN by creating an intermediary snapshot.

        @param server: the server to act on
        @param element: the storage element to clone
        @param clone_name: a name for the new clone.  If not provided,
            a name will be generated.
        @param qos_value: the QoS value to set on the storage element.
            This is the "value" field of a dict returned by
            get_qos_list().
        @param plugin_options: a string containing options specific
            to this plug-in
        @returns: the new clone element

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['create-clone', [element, clone_name, qos_value,
                              snapshot_element, plugin_options]],
        )
        return self._run_ops(**kwargs)

    def remove_element(self, server, element, **kwargs):
        """Remove a storage element from the server

        Removes a storage element and destroys all of its data.
        Storage elements can only be removed if they have no child
        clones or snapshots, and they are not accessible by any host.

        @param server: the server to act on
        @param element: the storage element to remove

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['remove-element', [element]],
        )
        return self._run_ops(**kwargs)

    # XXX extra_info
    def create_lun(self, server, volume_group, name, size,
                   thin_provisioned=False, qos_value=None,
                   plugin_options=None, **kwargs):
        """Create a new LUN on the storage server

        Creates a toplevel storage element on the storage server.

        @param server: the server to act on
        @param volume_group: the volume to create the LUN in
        @param name: the name of the new LUN
        @param size: the size of the new LUN in bytes
        @param thin_provisioned: a boolean describing whether the LUN
            should be thin provisioned or have all of its storage
            pre-allocated.
        @param qos_value: the QoS value to set on the storage element.
            This is the "value" field of a dict returned by
            get_qos_list().
        @param plugin_options: a string containing options specific
            to this plug-in
        @returns: the new LUN element

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['create-lun', [volume_group, name, size, thin_provisioned,
                            qos_value, plugin_options]],
        )
        return self._run_ops(**kwargs)

    def resize_element(self, server, element, new_size, **kwargs):
        """Resize a storage element

        LUNs, clones, and filesystems can be resized.  The new size must
        fit inside the volume group.

        @param server: the server to act on
        @param element: the storage element to resize
        @param new_size: the new size in bytes

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['resize-element', [element, new_size]],
        )
        return self._run_ops(**kwargs)

    def split_element(self, server, element, **kwargs):
        """Detach a storage element from any parent or peer

        Snapshots and clones share storage with their parent or peer
        elements.  split_element() will perform a deep copy and
        separate this element from any shared data regions.

        @param server: the server to act on
        @param element the storage element to split

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['split-element', [element]],
        )
        return self._run_ops(**kwargs)

    def restore_snapshot(self, server, snapshot, element=None, **kwargs):
        """Restore the state saved in a snapshot

        A snapshot is a point-in-time read-only copy of one or more
        storage elements.  Restoring a snapshot means reverting the
        original elements to the state contained in the snapshot.

        Storage server snapshots can contain old state for one or more
        storage elements depending on how they were created and what
        the storage supports.  If no original element is specified,
        every element described in the snapshot will be rolled back.
        If an original element is specified, only that element will
        be rolled back.  If rolling back a specific element would
        silently roll back other elements, the operation fails.

        @param server: the server to act on
        @param snapshot: the snapshot element to roll back
        @param element: the specific element to be rolled back

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['restore-snapshot', [snapshot, element]],
        )
        return self._run_ops(**kwargs)

    def create_filesystem(self, server, backing_device, name, size,
                          qos_value=None, force=False, plugin_options=None,
                          **kwargs):
        """Create a new filesystem on the storage server

        Creates a filesystem on the storage server.

        @param server: the server to act on
        @param backing_device: the backing object(s) to create the
            filesystem on
        @param name: the name of the new filesystem
        @param size: the size of the new filesystem in bytes
        @param qos_value: the QoS value to set on the filesystem
            This is the "value" field of a dict returned by
            get_qos_list().
        @param force: should an existing filesystem be overwritten
        @param plugin_options: a string containing options specific
            to this plug-in
        @returns: the new filesystem

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['create-filesystem', [backing_device, name, size,
                                   qos_value, force, plugin_options]],
        )
        return self._run_ops(**kwargs)

    def destroy_filesystem(self, server, fs, **kwargs):
        """Destroy a filesystem on the storage server

        Destroys a filesystem on the storage server.  All data will be
        lost.

        @param server: the server to act on
        @param fs: the filesystem to destroy

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['destroy-filesystem', [fs]],
        )
        return self._run_ops(**kwargs)

    def list_filesystems(self, server, **kwargs):
        """Retrieve a list of all filesystems on this server

        @param server: the record describing the server to query
        @returns: the list of filesystems

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-filesystems', None],
        )
        return self._run_ops(**kwargs)

    def list_mountpoints(self, server, fs=None, **kwargs):
        """Retrieve a list of all mountpoints on this server

        Returns the mountpoints for a given filesystem or for all
        filesystems known to the plugin, depending on whether a specific
        filesystem was specified.

        @param server: the record describing the server to query
        @param fs: the filesystem to ask about
        @returns: the list of mountpoints

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-mountpoints', [fs]],
        )
        return self._run_ops(**kwargs)

    def create_file(self, server, fs, file_name, file_type, size,
                          sparse = True, force = False,
                          **kwargs):

        """Create a new file on the storage server

        Creates a file on the storage server.

        @param server: the server to act on
        @param fs: the filesystem where the file should be created on
        @param file_name: file to create
        @param file_type: type of file to create
        @param size: the size of the new file in bytes
        @param sparse: if the the file should be sparsely allocated.
        @param force: should an existing file be overwritten
        @param plugin_options: a string containing options specific
            to this plug-in
        @returns: the new file

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['create-file', [fs, file_name, file_type, size,
                             sparse, force]],
        )
        return self._run_ops(**kwargs)

    def destroy_file(self, server, fs, file, **kwargs):

        """Destroy the file on the storage server

        Destroys the file on the storage server.

        @param server: the server to act on
        @param fs: the filesystem the file is on
        @param file: file to destroy

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['destroy-file', [fs, file]],
        )
        return self._run_ops(**kwargs)

    def resize_file(self, server, fs, file, new_size, sparse=True, **kwargs):

        """Resize the file on the storage server

        Resize the file on the storage server.

        @param server: the server to act on
        @param fs: the filesystem the file is on
        @param file: files to resize
        @param new_size: new size for the file
        @param sparse: should be sparsely allocated
        @returns: update file with actual file size filled in

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['resize-file', [fs, file, new_size, sparse]],
        )
        return self._run_ops(**kwargs)

    def clone_file(self, server, fs, file, clone_name=None, **kwargs):

        """Clone the file on the storage server

        Clone the file on the storage server.

        @param server: the server to act on
        @param fs: the filesystem the file is on
        @param file: files to clone
        @param clone_name: name of the clone
        @returns: new clone

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['clone-file', [fs, file, clone_name]],
        )
        return self._run_ops(**kwargs)

    def list_files(self, server, fs, files, recursive = False, **kwargs):
        """Retrieve a list of files on this server

        @param server: the record describing the server to query
        @param fs: the filesystem the files are on
        @param files: the files to list
        @returns: the list of files

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-files', [fs, files, recursive]],
        )
        return self._run_ops(**kwargs)


    def device_teardown(self, server, element, **kwargs):
        """Retrieve a list of files on this server

        @param server: the record describing the server the device is on
        @param element: the storage element that will be torn down
        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['teardown-device', [element]],
        )
        return self._run_ops(**kwargs)

    def device_resize(self, server, element, **kwargs):
        """Retrieve a list of files on this server

        @param server: the record describing the server the device is on
        @param element: the storage element that will be torn down
        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['resize-device', [element]],
        )
        return self._run_ops(**kwargs)

    def get_asyncprogress(self, server, result_record, **kwargs):
        """Get proress from async enabled plugin call

        @param server: the record describing the server preforming the async operation
        @param result_record: async progress up until now
        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['get-progress', [result_record]],
        )
        return self._run_ops(**kwargs)

    def mount(self, server, fs, mountpoint, share_path=None,
              create_mountpoint=False, mount_options=[], **kwargs):
        """Mount a filesystem on the server

        Given a filesystem and a mountpoint, mount the filesystem with
        the given options.

        @param server: the server to mount on
        @param fs: the filesystem to mount
        @param mountpoint: where to mount the filesystem
        @param share_path: subdirectory of the filesystem to mount
        @param create_mountpoint: create the mountpoint if necessary
        @param mount_options: options for the mount process
        @returns: the updated filesystem

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['mount', [fs, mountpoint, share_path, create_mountpoint,
                       mount_options]],
        )
        return self._run_ops(**kwargs)

    def unmount(self, server, fs, mountpoint, destroy_mountpoint=False,
                **kwargs):
        """Mount a filesystem on the server

        Given a filesystem and a mountpoint, mount the filesystem with
        the given options.

        @param server: the server to unmount on
        @param fs: the filesystem to unmount
        @param mountpoint: where to unmount the filesystem from
        @param destroy_mountpoint: destroy the mountpoint if capable
        @returns: the updated filesystem

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['unmount', [fs, mountpoint, destroy_mountpoint]],
        )
        return self._run_ops(**kwargs)

    def list_backing_devices(self, server, **kwargs):
        """List the plugin-named devices on which a filesystem can be created

        When filesystem.backing-device.type is "plugin-specified", the
        get_backing_devices() call returns the available backing devices.
        The 'value' member of each backing device dict is passed to
        create_filesystem().

        @param server: the server to query
        @returns: a list of backing devices

        """
        self._set_ops(
            ['set-plugin', [self.plugin_name]],
            ['set-server', [server]],
            ['list-backing-devices', None],
        )
        return self._run_ops(**kwargs)

