"""
    Copyright (C) 2010 Oracle.  All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
"""

import sys
import os
import asyncore
import asynchat
from errno import EBADF

__all__ = ['StdinoutChat']

class ChatChannel(asyncore.file_dispatcher):
    """Forward reads and writes to a chat object.

    This base channel forwards reads and writes to the associated
    chat object.  It is not meant to instantiated directly.  It is used
    by the StdinChannel and StdoutChannel objects.

    """
    def __init__(self, chat, fd, map=None, compat_fd=False):
        if compat_fd:
            fd = self._fixup_compat_fd(fd)
        self.chat = chat
        asyncore.file_dispatcher.__init__(self, fd, map)
        #sys.stderr.write('Opened ChatChannel for %s\n' % self)

    def _fixup_compat_fd(self, fd=None):
        #
        # Older asyncore (python < 2.6) doesn't handle file objects,
        # only raw file descriptors.  And it doesn't dup() them.
        #
        if fd:
            try:
                fd = fd.fileno()
            except AttributeError:
                pass
            fd = os.dup(fd)
        return fd

    def handle_read(self):
        # Pass the read event on to our chat
        self.chat.handle_read()

    def handle_write(self):
        # Pass the read event on to our chat
        self.chat.handle_write()

    def handle_close(self):
        #sys.stderr.write('Close fired for %s\n' % self)
        self.close()

    def close(self):
        asyncore.file_dispatcher.close(self)
        #sys.stderr.write('Closed ChatChannel for %s\n' % self)
        self.chat = None

    def log(self, message):
        pass

    def log_info(self, message, type='info'):
        pass


class StdinChannel(ChatChannel):
    """A channel that passes stdin events to the associated chat."""
    def __init__(self, chat, stdin_fd=None, map=None, compat_fd=False):
        if not stdin_fd:
            stdin_fd = sys.stdin
        ChatChannel.__init__(self, chat, stdin_fd, map, compat_fd)

    def handle_close(self):
        self.chat.handle_in_close()
        ChatChannel.handle_close(self)

    def writable(self):
        return 0

    def readable(self):
        return self.chat and self.chat.readable()


class StdoutChannel(ChatChannel):
    """A channel that passes stdout events to the associated chat."""
    def __init__(self, chat, stdout_fd=None, map=None, compat_fd=False):
        if not stdout_fd:
            stdout_fd = sys.stdout
        ChatChannel.__init__(self, chat, stdout_fd, map, compat_fd)

    def handle_close(self):
        self.chat.handle_out_close()
        ChatChannel.handle_close(self)

    def writable(self):
        return self.chat and self.chat.writable()

    def readable(self):
        return 0


class StdinoutChat(asynchat.async_chat):
    """async_chat over standard input and output.

    This is a version of async_chat that takes its input on stdin and
    pushes output to stdout.  It does this by creating proxy channels.
    The proxy channels call handle_read() and handle_write() here.  In
    return, this chat uses the proxy channels' send() and recv()
    methods.  Alternate file objects or descriptors can be provided to
    support pipes.

    Like async_chat, this still needs to be subclassed.

    """
    def __init__(self, stdin_fd=None, stdout_fd=None, map=None):
        compat_fd = False
        try:
            asynchat.async_chat.__init__(self, None, map)
        except TypeError, e:
            asynchat.async_chat.__init__(self, None)
            if map:
                self._map = map
            #
            # Passing compat_fd all the way down is a hack, but this is
            # the place where we know it is needed.
            #
            compat_fd = True
        self.connected = True
        self.in_channel = StdinChannel(self, stdin_fd, map, compat_fd)
        self.out_channel = StdoutChannel(self, stdout_fd, map, compat_fd)

    def send(self, *args):
        return self.out_channel.send(*args)

    def recv(self, *args):
        return self.in_channel.recv(*args)

    def handle_in_close(self):
        self.close_when_done()

    def handle_out_close(self):
        self.handle_close()

    def add_channel(self, map=None):
        self.in_channel.add_channel(map)
        self.out_channel.add_channel(map)

    def del_channel(self, map=None):
        self.in_channel.del_channel(map)
        self.out_channel.del_channel(map)

    def close(self):
        #sys.stderr.write('close fired for %s\n' % repr(self))
        for chan in [self.in_channel, self.out_channel]:
            try:
                chan.close()
            except OSError, e:
                if e.args[0] != EBADF:
                    raise
        self.in_channel = None
        self.out_channel = None

    def get_fds(self):
        if self.in_channel:
            return (self.in_channel.socket.fileno(),
                    self.out_channel.socket.fileno())
