"""
    Copyright (C) 2010 Oracle.  All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
"""

import sys

from OSCPlugin import ABILITY_TYPES, BACKING_DEVICE_TYPES, IPlugin
from sccaller.caller import SCCaller

__all__ = ['SCCallerOVM3']


#
# if_() utility function from:
# Peter Norvig's python utilities.
# http://aima-python.googlecode.com/svn/trunk/utils.py
# MIT License
#
def if_(test, result, alternative):
    """Like C++ and Java's (test ? result : alternative), except
    both result and alternative are always evaluated. However, if
    either evaluates to a function, it is applied to the empty arglist,
    so you can delay execution by putting it in a lambda.
    >>> if_(2 + 2 == 4, 'ok', lambda: expensive_computation())
    'ok'
    """
    if test:
        if callable(result): return result()
        return result
    else:
        if callable(alternative): return alternative()
        return alternative


class SCCallerOVM3(SCCaller):
    def __init__(self, plugin_name, logging_level=0, cache_notifier=None):
        """Access a Storage Connect plug-in with the OVM3 API.

        Instantiates a new caller to access a Storage Connect plug-in
        using the OVM3 plugin-in API.

        @param plugin_name: The plugin this caller will use

        """
        SCCaller.__init__(self, plugin_name, logging_level, cache_notifier)

    def _get_simple_async_hooks(self):
        ret = (self._ovm3_finish, self._ovm3_callback)
        self._ovm3_finish = None
        self._ovm3_callback = None
        return ret

    def _set_simple_async_hooks(self, finish, callback):
        self._ovm3_finish = finish
        self._ovm3_callback = callback

    def _simple_async_callback(self, result, e):
        finish, callback = self._get_simple_async_hooks()
        user_result = if_(e, None, lambda : finish(result))
        callback(user_result, e)

    def _setup_simple_async(self, kwargs, finish_func):
        """Handle call proxying with simple result conversion

        Many OVM3 calls do not return exactly what the underlying
        SCCaller API returns.  They have to convert the SCCaller result
        to the expected OVM3 answer.  For an asynchronous operation,
        this conversion has to happen in a callback.

        The simplest version makes a single SCCaller call and then
        tweaks the result.  If the call is synchronous, OVM3 can merely
        perform the conversion ("finish" the call) and return the
        result.  However, an asynchronous operation must take the
        SCCaller callback, perform the conversion, and then call the
        user's callback.

        This method uses the kwargs from the original call to
        determine whether the user requested synchronous or
        asynchronous operation.  If the operation is asynchronous, it
        will override the user's callback with
        _simple_async_callback().  That function will perform the
        conversion by passing the result as single argument to
        finish_func() and call the user's callback.

        The calling method takes the function returned from this method
        and applies it to the result of the SCCaller call.  If the mode
        is asynchronous, that result will be a (read_fd, write_fd) tuple
        the functioned returned will pass that along.  If the mode is
        synchronous, this method will return the finish_func(), which
        will convert the result appropriately.

        An example.  Let's say you have a SCCaller method named
        'last_index()'.  It returns the last index of a list.  However,
        the OVM3 returns the count, not the last index:

        >>> SCCaller.last_index([1, 2, 3])
        2
        >>> SCCallerOVM3.count([1, 2, 3])
        3

        A simple synchronous OVM3 method might looks like this:

        >>> def count(self, l):
        >>>     return self.last_index(l) + 1

        The asynchronous-capable version looks like this:

        >>> def _count_finish(self, index):
        >>>     return index + 1

        >>> def count(self, l, **kwargs):
        >>>     finish = self._setup_simple_async(kwargs,
        >>>                                       self._count_finish)
        >>>     result = self.last_index(l, **kwargs)
        >>>     return finish(result)

        @param kwargs: the keyword arguments passed to the OVM3 method
        @param finish_func: the conversion function used to process the
            SCCaller method's result.
        @returns: a finish function appropriate for the mode of
            operation.  The calling method should apply this function
            to the SCCaller method's initial result and return it to
            the user.

        """
        if kwargs.has_key('callback'):
            self._set_simple_async_hooks(finish_func, kwargs['callback'])
            kwargs['callback'] = self._simple_async_callback
            return lambda x: x
        else:
            return finish_func

    def discoverPlugins(self, **kwargs):
        return self.list_plugins(**kwargs)

    def _map_attributes(self, attributes):
        """Convert attributes into the plugin info dict of the OVM3 API.

        Convert attributes returned by get-plugin-attributes into the
        info dict expected by the OVM3 API.  This is the reverse of the
        mapping done by the worker.

        """
        info_map_generic = {
            'name': 'plugin_name', 'type': 'plugin_type',
            'vendor': 'vendor_name', 'description': 'plugin_desc',
            'api-version': 'required_api_vers',
            'help.extra-info.server': 'ss_extra_info_help',
        }
        info_map_storage = {
            'storage.types': 'storage_types',
            'storage.api-version': 'sa_api_version',
            'help.extra-info.element': 'se_extra_info_help',
        }
        info_map_filesystem = {
            'filesystem.name': 'filesys_name',
            'filesystem.type': 'filesys_type',
            'filesystem.api-version': 'fs_api_version',
            'help.extra-info.filesystem': 'fs_extra_info_help',
            'help.extra-info.file': 'file_extra_info_help',
        }

        working_map = info_map_generic.copy()
        if attributes['type'] == 'ifs':
            working_map.update(info_map_filesystem)
        elif attributes['type'] == 'isa':
            working_map.update(info_map_storage)
        else:
            raise ValueError('Plugin %s is of an unknown plugin type: "%s"' %
                             (self.plugin_name, attributes['type']))
        info = {}
        for name, info_name in working_map.iteritems():
            info[info_name] = attributes[name]
        return info

    def _map_backing_device_type(self, attributes):
        if not attributes.has_key('filesystem.backing-device.type'):
            return None
        if attributes['filesystem.backing-device.type'] == 'device':
            if attributes['filesystem.backing-device.multi']:
                return BACKING_DEVICE_TYPES.DEVICE_MULTI
            else:
                return BACKING_DEVICE_TYPES.DEVICE_SINGLE
        elif attributes['filesystem.backing-device.type'] == 'plugin-specified':
            if attributes['filesystem.backing-device.multi']:
                return BACKING_DEVICE_TYPES.PLUGIN_MULTI
            else:
                return BACKING_DEVICE_TYPES.PLUGIN_SINGLE
        elif attributes['filesystem.backing-device.type'] == 'unsupported':
            return BACKING_DEVICE_TYPES.UNSUPPORTED
        else:
            raise ValueError('Plugin %s has an unknown backing device type: "%s"' %
                             (self.plugin_name,
                              attributes['filesystem.backing-device.type']))

    def _map_capabilities(self, attributes):
        """Convert capability attributes to an OVM3 ability dict.

        Convert capability attributes returned by get-plugin-attributes or
        get-server-attributes into the capabilities dict expected by the
        OVM3 API.  This is the reverse of the mapping done by the worker.

        """
        # Use if_() to compute ABILITY values.
        ability_p = \
            lambda op: if_(attributes['capability.%s' % op],
                           ABILITY_TYPES.YES, ABILITY_TYPES.NO)
        ability_not_p = \
            lambda op: if_(attributes['capability.%s' % op],
                           ABILITY_TYPES.NO, ABILITY_TYPES.YES)
        # Use if_() to compute ONLINE/OFFLINE/NO ability values.
        ability_online_p = \
            lambda op: if_(attributes['capability.%s' % op],
                           if_(attributes['capability.%s.online' % op],
                               ABILITY_TYPES.ONLINE, ABILITY_TYPES.OFFLINE),
                           ABILITY_TYPES.NO)
        plugin_ability = {
            'snapshot':         ability_online_p('snapshot'),
            'custom_snap_name': ability_p('snapshot.custom-name'),
            'snap_is_sync':     ability_not_p('snapshot.asynchronous'),
            'clone':            ability_online_p('clone'),
            'custom_clone_name': ability_p('clone.custom-name'),
            'clone_is_sync':    ability_not_p('clone.asynchronous'),
            'resize':           ability_online_p('resize'),
            'resize_is_sync':   ability_not_p('resize.asynchronous'),
            'splitclone':       ability_online_p('splitclone'),
            'splitclone_is_sync': ability_not_p('splitclone.asynchronous'),
            'splitclone_while_open': ability_p('splitclone.open'),
            'snapclone':        ability_online_p('snapclone'),
            'snapclone_is_sync': ability_not_p('snapclone.asynchronous'),
            'require_storage_name': ability_p('storage-name-required'),
            'access_control':   ability_p('access-control'),
            'max_access_entries':
                attributes['capability.access-control.max-entries'],
        }
        backing_type = self._map_backing_device_type(attributes)
        if backing_type:
            plugin_ability['backing_device_type'] = backing_type

        return plugin_ability

    def _getPluginInfo_finish(self, attributes):
        info = self._map_attributes(attributes)
        info['plugin_ability'] = self._map_capabilities(attributes)
        return info

    def getPluginInfo(self, **kwargs):
        finish = self._setup_simple_async(kwargs,
                                          self._getPluginInfo_finish)
        result = self.get_plugin_attributes('*', **kwargs)
        return finish(result)

    def validate(self, ss_record, **kwargs):
        return SCCaller.validate(self, ss_record, **kwargs)

    def _getCapabilities_finish(self, attributes):
        return self._map_capabilities(attributes)

    def getCapabilities(self, ss_record, **kwargs):
        finish = self._setup_simple_async(kwargs,
                                          self._getCapabilities_finish)
        result = self.get_server_attributes(ss_record, '*', **kwargs)
        return finish(result)

    def getInfo(self, ss_record, se_record=None, **kwargs):
        if se_record:
            return self.get_element_info(ss_record, se_record, **kwargs)
        else:
            return self.get_server_info(ss_record, **kwargs)

    def getStorageServerInfo(self, ss_record, **kwargs):
        return self.get_server_info(ss_record, **kwargs)

    def getFileSystemInfo(self, ss_record, fs_record, **kwargs):
        return self.get_element_info(ss_record, fs_record, **kwargs)

    def getStorageNames(self, ss_record, **kwargs):
        return self.get_storage_names(ss_record, **kwargs)

    def getStatus(self, ss_record, obj_record=None, **kwargs):
        if obj_record:
            return self.get_element_status(ss_record, obj_record, **kwargs)
        else:
            return self.get_server_status(ss_record, **kwargs)

    def _list_finish(self, e):
        self._multi_op_end()
        result = if_(e, None, lambda : self._ovm3_list['result'])
        callback = self._ovm3_callback
        self._ovm3_callback = None
        del self._ovm3_list
        callback(result, e)

    def _list_callback(self, result, e):
        if e:
            self._list_finish(e)
        self._ovm3_list['result'].extend(result)
        self._list_run()

    def _list_run(self):
        if len(self._ovm3_list['se_type']):
            next_se_type = self._ovm3_list['se_type'].pop(0)
            if next_se_type == IPlugin.LUNType:
                return self.list_luns(self._ovm3_list['ss'],
                                      **self._ovm3_list['kwargs'])
            elif next_se_type == IPlugin.SnapCloneType:
                return self.list_clones(self._ovm3_list['ss'],
                                        **self._ovm3_list['kwargs'])
        self._list_finish(None)

    def _list_async(self, ss_record, se_type, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._list_callback
        self._ovm3_list = {
            'ss': ss_record.copy(),
            'se_type': se_type[:],
            'kwargs': kwargs,
            'result': [],
        }
        self._multi_op_start()
        return self._list_run()

    def list(self, ss_record, se_type=[IPlugin.LUNType, IPlugin.SnapCloneType],
             **kwargs):
        if kwargs.has_key('callback'):
            return self._list_async(ss_record, se_type, **kwargs)
        self._multi_op_start()
        records = []
        if IPlugin.LUNType in se_type:
            records.extend(self.list_luns(ss_record, **kwargs))
        if IPlugin.SnapCloneType in se_type:
            records.extend(self.list_clones(ss_record, **kwargs))
        self._multi_op_end()
        return records

    def _create_file_finish(self, e):
        self._multi_op_end()
        result = if_(e, None, lambda : self._ovm3_create_file['result'])
        callback = self._ovm3_callback
        self._ovm3_callback = None
        callback(result, e)

    def _create_file_callback(self, result, e):
        if e:
            self._create_file_finish(e)
        self._ovm3_create_file['result'].extend(result)
        self._create_file_finish(None)

    def _create_file_async(self, ss_record, file_name,
                                 file_type, size, sparse, force,
                                 **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._create_file_callback
        self._multi_op_start()
        return self.create_file(ss_record, fs_record, file_name,
                                file_type, size, sparse, force,
                                **kwargs)

    def createFile(self, ss_record, fs_record, file_name, file_type,
                         size, sparse=True, force=False, **kwargs):
        if kwargs.has_key('callback'):
            return self._create_file_async(ss_record, fs_record, file_name,
                                           file_type, size, sparse, force,
                                           **kwargs)
        return self.create_file(ss_record, fs_record, file_name, file_type,
                                size, sparse, force, **kwargs)

    def resizeFile(self, ss_record, fs_record, file_record, new_size,
                         sparse=True, **kwargs):
        return self.resize_file(ss_record, fs_record, file_record,
                                new_size, **kwargs)

    def _clone_file_finish(self, e):
        self._multi_op_end()
        result = if_(e, None, lambda : self._ovm3_clone_file['result'])
        callback = self._ovm3_callback
        self._ovm3_callback = None
        callback(result, e)

    def _clone_file_callback(self, result, e):
        if e:
            self._create_file_finish(e)
        self._ovm3_clone_file['result'].extend(result)
        self._clone_file_finish(None)

    def _clone_file_async(self, ss_record, fs_record, file_record,
                                dest_file_name, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._clone_file_callback
        self._multi_op_start()
        return self.clone_file(ss_record, fs_record, file_record,
                              dest_file_name, **kwargs)
    
    def cloneFile(self, ss_record, fs_record, file_record,
                        dest_file_name=None, **kwargs):
        if kwargs.has_key('callback'):
            return self._clone_file_async(ss_record, fs_record, file_record,
                                          dest_file_name, **kwargs)
        return self.clone_file(ss_record, fs_record, file_record,
                               dest_file_name, **kwargs)

    def destroyFile(self, ss_record, fs_record, file_record, **kwargs):
        return self.destroy_file(ss_record, fs_record, file_record,
                                 **kwargs)

    def _list_files_finish(self, e):
        self._multi_op_end()
        result = if_(e, None, lambda : self._ovm3_list_files['result'])
        callback = self._ovm3_callback
        self._ovm3_callback = None
        callback(result, e)

    def _list_files_callback(self, result, e):
        if e:
            self._list_files_finish(e)
        self._ovm3_list_files['result'].extend(result)
        self._list_files_finish(None)

    def _list_files_async(self, ss_record, fs_record, file_record, recursive,
                          **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._list_files_callback
        self._multi_op_start()
        return self.list_files(ss_record, fs_record, file_record, recursive,
                               **kwargs)

    def listFiles(self, ss_record, fs_record, file_record, recursive=False,
             **kwargs):
        if kwargs.has_key('callback'):
            return self._list_files_async(ss_record, fs_record, file_record, 
                                          recursive, **kwargs)
        return self.list_files(ss_record, fs_record, file_record, recursive,
                               **kwargs)

    def deviceTeardown(self, ss_record, se_record, **kwargs):
        return self.device_teardown(ss_record, se_record, **kwargs)

    def deviceResize(self, ss_record, se_record, **kwargs):
        return self.device_resize(ss_record, se_record, **kwargs)

    def getCurrentClones(self, ss_record, se_record=None, **kwargs):
        if se_record:
            return self.list_clones(ss_record, se_record, **kwargs)
        else:
            return self.list_clones(ss_record, **kwargs)

    def getCurrentSnaps(self, ss_record, se_record=None, **kwargs):
        if se_record:
            return self.list_snapshots(ss_record, se_record, **kwargs)
        else:
            return self.list_snapshots(ss_record, **kwargs)

    def _isCloneable_finish(self, attributes):
        return attributes['capability.clone']

    def isCloneable(self, ss_record, se_record, **kwargs):
        finish = self._setup_simple_async(kwargs, self._isCloneable_finish)
        result = self.get_element_attributes(ss_record, se_record,
                                             'capability.clone',
                                             **kwargs)
        return finish(result)

    def _isSnapable_finish(self, attributes):
        return attributes['capability.snapshot']

    def isSnapable(self, ss_record, se_record, **kwargs):
        finish = self._setup_simple_async(kwargs, self._isSnapable_finish)
        result = self.get_element_attributes(ss_record, se_record,
                                             'capability.snapshot',
                                             **kwargs)
        return finish(result)

    def _isRestorable_finish(self, attributes):
        return attributes['capability.restore']

    def isRestorable(self, ss_record, se_record, **kwargs):
        finish = self._setup_simple_async(kwargs, self._isRestorable_finish)
        result = self.get_element_attributes(ss_record, se_record,
                                             'capability.restore',
                                             **kwargs)
        return finish(result)

    def _isSplittable_finish(self, attributes):
        return attributes['capability.split']

    def isSplittable(self, ss_record, se_record, **kwargs):
        finish = self._setup_simple_async(kwargs, self._isSplittable_finish)
        result = self.get_element_attributes(ss_record, se_record,
                                             'capability.split', **kwargs)
        return finish(result)

    def _getCloneLimits_finish(self, attributes):
        return attributes['capability.clone.max']

    def _getCloneLimits(self, ss_record, obj_record=None, **kwargs):
        finish = self._setup_simple_async(kwargs,
                                          self._getCloneLimits_finish)
        if obj_record:
            result = self.get_element_attributes(ss_record, obj_record,
                                                 'capability.clone.max',
                                                 **kwargs)
        else:
            result = self.get_server_attributes(ss_record,
                                                'capability.clone.max',
                                                **kwargs)
        return finish(result)

    def getCloneLimits(self, ss_record, se_record=None, **kwargs):
        return self._getCloneLimits(ss_record, se_record, **kwargs)

    def getFileSystemCloneLimits(self, ss_record, fs_record=None, **kwargs):
        return self._getCloneLimits(ss_record, fs_record, **kwargs)

    def _getSnapLimits_finish(self, attributes):
        return attributes['capability.snapshot.max']

    def _getSnapLimits(self, ss_record, obj_record=None, **kwargs):
        finish = self._setup_simple_async(kwargs,
                                          self._getSnapLimits_finish)
        if obj_record:
            result = self.get_element_attributes(ss_record, obj_record,
                                                 'capability.snapshot.max',
                                                 **kwargs)
        else:
            result = self.get_server_attributes(ss_record,
                                                'capability.snapshot.max',
                                                **kwargs)
        return finish(result)

    def getSnapLimits(self, ss_record, se_record=None, **kwargs):
        return self._getSnapLimits(ss_record, se_record, **kwargs)

    def getFileSystemSnapLimits(self, ss_record, fs_record=None, **kwargs):
        return self._getSnapLimits(ss_record, fs_record, **kwargs)

    def getAccessGroups(self, ss_record, se_record=None, **kwargs):
        if se_record:
            return self.list_access_groups(ss_record, se_record, **kwargs)
        else:
            return self.list_access_groups(ss_record, **kwargs)

    def _ag_finish(self, ags, e):
        self._multi_op_end()
        if not e:
            ss_record = self._ovm3_ag['ss']
            ss_record['access_grps'] = ags
        else:
            ss_record = None
        callback = self._ovm3_callback
        self._ovm3_callback = None
        del self._ovm3_ag
        callback(ss_record, e)

    def _ag_callback(self, result, e):
        if e:
            self._ag_finish(None, e)
        self._ag_run()

    def _ag_run(self):
        kwargs = self._ovm3_ag['kwargs']
        if len(self._ovm3_ag['access_groups']):
            ag = self._ovm3_ag['access_groups'].pop(0)
            return self._ovm3_ag['function'](self._ovm3_ag['ss'],
                                             ag['grp_name'],
                                             *ag['grp_entries'], **kwargs)
        kwargs['callback'] = self._ag_finish
        return self.list_access_groups(self._ovm3_ag['ss'], **kwargs)

    def _ag_async(self, mkrm_func, ss_record, access_groups, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._ag_callback
        self._ovm3_ag = {
            'ss': ss_record.copy(),
            'function': mkrm_func,
            'access_groups': access_groups[:],
            'kwargs': kwargs,
        }
        self._multi_op_start()
        return self._ag_run()

    def createAccessGroups(self, ss_record, access_groups, **kwargs):
        if kwargs.has_key('callback'):
            return self._ag_async(self.create_access_group, ss_record,
                                  access_groups, **kwargs)
        self._multi_op_start()
        for ag in access_groups:
            self.create_access_group(ss_record, ag['grp_name'],
                                     *ag['grp_entries'])
        ags = self.list_access_groups(ss_record)
        ss_record['access_grps'] = ags
        self._multi_op_end()
        return ss_record

    def removeAccessGroups(self, ss_record, access_groups, **kwargs):
        if kwargs.has_key('callback'):
            return self._ag_async(self.remove_access_group, ss_record,
                                  access_groups, **kwargs)
        self._multi_op_start()
        for ag in access_groups:
            self.remove_access_group(ss_record, ag['grp_name'])
        ags = self.list_access_groups(ss_record)
        ss_record['access_grps'] = ags
        self._multi_op_end()
        return ss_record

    def _rename_access_group_callback(self, result, e):
        if e:
            self._ag_finish(None, e)
        kwargs = self._ovm3_ag['kwargs']
        kwargs['callback'] = self._ag_finish
        return self.list_access_groups(self._ovm3_ag['ss'], **kwargs)

    def _renameAccessGroup_async(self, ss_record, access_group_name,
                                 new_access_group_name, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._rename_access_group_callback
        self._ovm3_ag = {
            'ss': ss_record.copy(),
            'kwargs': kwargs,
        }
        self._multi_op_start()
        return self.rename_access_group(ss_record, access_group_name,
                                        new_access_group_name, **kwargs)

    def renameAccessGroup(self, ss_record, access_group_name,
                          new_access_group_name, **kwargs):
        if kwargs.has_key('callback'):
            return self._renameAccessGroup_async(ss_record,
                                                 access_group_name,
                                                 new_access_group_name,
                                                 **kwargs)
        self._multi_op_start()
        self.rename_access_group(ss_record, access_group_name,
                                 new_access_group_name)
        ags = self.list_access_groups(ss_record)
        self._multi_op_end()
        ss_record['access_grps'] = ags
        return ss_record

    def _ag_host_finish(self, ags, e):
        self._multi_op_end()
        ag = None
        if not e:
            for try_ag in ags:
                if try_ag['grp_name'] == self._ovm3_ag['ag_name']:
                    ag = try_ag
        callback = self._ovm3_callback
        self._ovm3_callback = None
        del self._ovm3_ag
        callback(ag, e)

    def _ag_host_callback(self, result, e):
        if e:
            self._ag_host_finish(None, e)
        kwargs = self._ovm3_ag['kwargs']
        kwargs['callback'] = self._ag_host_finish
        return self.list_access_groups(self._ovm3_ag['ss'], **kwargs)

    def _ag_host_async(self, add_remove_func, ss_record, access_group_name,
                       entries, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._ag_host_callback
        self._ovm3_ag = {
            'ss': ss_record.copy(),
            'ag_name': access_group_name,
            'kwargs': kwargs,
        }
        self._multi_op_start()
        return add_remove_func(ss_record, access_group_name, *entries,
                               **kwargs)

    def addToAccessGroup(self, ss_record, access_group_name, entries,
                         **kwargs):
        if kwargs.has_key('callback'):
            return self._ag_host_async(self.add_hosts_to_access_group,
                                       ss_record, access_group_name,
                                       entries, **kwargs)
        self._multi_op_start()
        self.add_hosts_to_access_group(ss_record, access_group_name,
                                       *entries)
        ags = self.list_access_groups(ss_record)
        self._multi_op_end()
        for ag in ags:
            if ag['grp_name'] == access_group_name:
                return ag

    def removeFromAccessGroup(self, ss_record, access_group_name, entries,
                              **kwargs):
        if kwargs.has_key('callback'):
            return self._ag_host_async(self.remove_hosts_from_access_group,
                                       ss_record, access_group_name,
                                       entries, **kwargs)
        self._multi_op_start()
        self.remove_hosts_from_access_group(ss_record, access_group_name,
                                            *entries)
        ags = self.list_access_groups(ss_record)
        self._multi_op_end()
        for ag in ags:
            if ag['grp_name'] == access_group_name:
                return ag

    def _iscsi_sendtargets_callback(self, targets, e, ss_record, callback):
        if e:
            callback(None, e)
        ss_record['storage_id'] = targets
        callback(ss_record, None)

    def discover(self, ss_record, **kwargs):
        if kwargs.has_key('callback'):
            callback = kwargs['callback']
            kwargs['callback'] = \
                lambda result, e: self._iscsi_sendtargets_callback(result,
                                                                   e,
                                                                   ss_record,
                                                                   callback)
            return self.iscsi_sendtargets(ss_record, **kwargs)
        targets = self.iscsi_sendtargets(ss_record)
        ss_record['storage_id'] = targets
        return ss_record

    def start(self, ss_record, **kwargs):
        targets = ss_record.get('storage_id', [])
        return self.iscsi_start(ss_record, *targets, **kwargs)

    def stop(self, ss_record, **kwargs):
        targets = ss_record.get('storage_id', [])
        return self.iscsi_stop(ss_record, *targets, **kwargs)

    def refresh(self, ss_record, **kwargs):
        targets = ss_record.get('storage_id', [])
        return self.lun_scan(ss_record, *targets, **kwargs)

    def _map_finish(self, e):
        self._multi_op_end()
        result = if_(e, None, lambda : self._ovm3_map['mapped'])
        callback = self._ovm3_callback
        self._ovm3_callback = None
        del self._ovm3_map
        callback(result, e)

    def _map_callback(self, paths, e):
        if e:
            self._map_finish(e)
        se_record = self._ovm3_map.pop('mapping')
        se_record['paths'] = paths
        self._ovm3_map['mapped'].append(se_record)
        self._map_run()

    def _map_run(self):
        if len(self._ovm3_map['to_map']):
            self._ovm3_map['mapping'] = self._ovm3_map['to_map'].pop(0)
            return self.map_devices(self._ovm3_map['ss'],
                                    self._ovm3_map['mapping'],
                                    **self._ovm3_map['kwargs'])
        return self._map_finish(None)

    def _map_async(self, ss_record, se_records, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._map_callback
        self._ovm3_map = {
            'ss': ss_record.copy(),
            'to_map': se_records[:],
            'mapped': [],
            'kwargs': kwargs,
        }
        self._multi_op_start()
        return self._map_run()

    def updateSERecords(self, ss_record, se_records, append_allowed=False,
                        **kwargs):
        if kwargs.has_key('callback'):
            return self._map_async(ss_record, se_records, **kwargs)
        self._multi_op_start()
        for rec in se_records:
            rec['paths'] = self.map_devices(ss_record, rec)
        self._multi_op_end()
        return se_records

    def online(self, ss_record, se_record, **kwargs):
        return self.element_online(ss_record, se_record, **kwargs)

    def offline(self, ss_record, se_record, **kwargs):
        return self.element_offline(ss_record, se_record, **kwargs)

    def getQoSList(self, ss_record, element=None, **kwargs):
        return self.get_qos_list(ss_record, element, **kwargs)

    def setQoS(self, ss_record, se_record, qos_value, **kwargs):
        return self.set_element_qos(ss_record, se_record, qos_value,
                                    **kwargs)

    def createSnap(self, ss_record, se_record, snap_se_record=None,
                   **kwargs):
        name = if_(snap_se_record, lambda :snap_se_record['name'], None)
        return self.create_snapshot(ss_record, se_record, name,
                                    **kwargs)

    def createMultiSnap(self, ss_record, se_records, snap_se_record=None,
                        **kwargs):
        name = if_(snap_se_record, lambda :snap_se_record['name'], None)
        return self.create_multi_element_snapshot(ss_record, se_records,
                                                  name, **kwargs)

    def clone(self, ss_record, se_record, dest_se_record=None, qos=None,
              **kwargs):
        name = if_(dest_se_record, lambda :dest_se_record['name'], None)
        plugin_options = if_(dest_se_record,
                             lambda :dest_se_record.get('extra_info', None),
                             None)
        return self.create_clone(ss_record, se_record, name, qos,
                                 plugin_options, **kwargs)

    def cloneFromSnap(self, ss_record, se_record, snap_se_record,
                      dest_se_record=None, qos=None, **kwargs):
        name = if_(dest_se_record, lambda :dest_se_record['name'], None)
        plugin_options = if_(dest_se_record,
                             lambda :dest_se_record.get('extra_info', None),
                             None)
        return self.create_clone(ss_record, se_record, name, qos,
                                 snap_se_record, plugin_options, **kwargs)

    def remove(self, ss_record, se_record, **kwargs):
        return self.remove_element(ss_record, se_record, **kwargs)

    def snapRemove(self, ss_record, snap_se_record, **kwargs):
        return self.remove_element(ss_record, snap_se_record, **kwargs)

    def create(self, ss_record, se_record, thin_provision=True, qos=None,
               **kwargs):
        return self.create_lun(ss_record, se_record['vol_group_name'],
                               se_record['name'], se_record['size'],
                               thin_provision, qos,
                               se_record.get('extra_info', None), **kwargs)


    def _resize_finish(self, lun, e):
        self._multi_op_end()
        callback = self._ovm3_callback
        self._ovm3_callback = None
        del self._ovm3_resize
        callback(lun, e)

    def _resize_callback(self, result, e):
        if e:
            self._resize_finish(None, e)
        kwargs = self._ovm3_resize['kwargs']
        kwargs['callback'] = self._resize_finish
        return self.get_element_info(self._ovm3_resize['ss'],
                                     self._ovm3_resize['se'], **kwargs)

    def _resize_async(self, ss_record, se_record, new_size, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._resize_callback
        self._ovm3_resize = {
            'ss': ss_record.copy(),
            'se': se_record.copy(),
            'kwargs': kwargs,
        }
        self._multi_op_start()
        return self.resize_element(ss_record, se_record, new_size, **kwargs)

    def resize(self, ss_record, se_record, new_size, **kwargs):
        if kwargs.has_key('callback'):
            return self._resize_async(ss_record, se_record, new_size,
                                      **kwargs)
        self._multi_op_start()
        self.resize_element(ss_record, se_record, new_size)
        lun = self.get_element_info(ss_record, se_record)
        self._multi_op_end()
        return lun

    def _present_finish(self, lun, e):
        self._multi_op_end()
        callback = self._ovm3_callback
        self._ovm3_callback = None
        del self._ovm3_present
        callback(lun, e)

    def _present_callback(self, result, e):
        if e:
            self._present_finish(None, e)
            return
        self._present_run()

    def _present_run(self):
        kwargs = self._ovm3_present['kwargs']
        if len(self._ovm3_present['access_group_names']):
            name = self._ovm3_present['access_group_names'].pop(0)
            return self._ovm3_present['function'](self._ovm3_present['ss'],
                                                  name,
                                                  self._ovm3_present['se'],
                                                  **kwargs)
        kwargs['callback'] = self._present_finish
        return self.get_element_info(self._ovm3_present['ss'],
                                     self._ovm3_present['se'], **kwargs)

    def _present_async(self, present_func, ss_record, se_record,
                       access_group_names, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._present_callback
        self._ovm3_present = {
            'ss': ss_record.copy(),
            'se': se_record.copy(),
            'function': present_func,
            'access_group_names': access_group_names[:],
            'kwargs': kwargs,
        }
        return self._present_run()

    def startPresent(self, ss_record, se_record, access_grp_names,
                     **kwargs):
        if kwargs.has_key('callback'):
            self._multi_op_start()
            return self._present_async(self.add_elements_to_access_group,
                                       ss_record, se_record,
                                       access_grp_names, **kwargs)
        self._multi_op_start()
        for ag_name in access_grp_names:
            self.add_elements_to_access_group(ss_record, ag_name,
                                              se_record)
        lun = self.get_element_info(ss_record, se_record)
        self._multi_op_end()
        return lun

    def stopPresent(self, ss_record, se_record, access_grp_names, **kwargs):
        if kwargs.has_key('callback'):
            self._multi_op_start()
            return self._present_async(self.remove_elements_from_access_group,
                                       ss_record, se_record,
                                       access_grp_names, **kwargs)
        self._multi_op_start()
        for ag_name in access_grp_names:
            self.remove_elements_from_access_group(ss_record, ag_name,
                                                   se_record)
        lun = self.get_element_info(ss_record, se_record)
        self._multi_op_end()
        return lun

    
    def splitClone(self, ss_record, se_record, **kwargs):
        return self.split_element(ss_record, se_record, **kwargs)


    def snapRestore(self, ss_record, snap_se_record, se_record=None,
                    **kwargs):
        return self.restore_snapshot(ss_record, snap_se_record, se_record,
                                     **kwargs)

    def getAsyncProgress(self, ss_record, result_record, **kwargs):
        return self.get_asyncprogress(ss_record, result_record, **kwargs)

    def _mkfs_callback(self, result, e):
        if e:
            self._ovm3_present = self._ovm3_mkfs
            del self._ovm3_mkfs
            self._present_finish(None, e)
            return
        ss_record = self._ovm3_mkfs['ss']
        access_grp_names = self._ovm3_mkfs['access_group_names']
        kwargs = self._ovm3_mkfs['kwargs']
        kwargs['callback'] = self._ovm3_callback
        self._ovm3_callback = None
        del self._ovm3_mkfs
        return self._present_async(self.add_elements_to_access_group,
                                   ss_record, result, access_grp_names,
                                   **kwargs)

    def _mkfs_async(self, ss_record, name, backing_device, size,
                    access_grp_names, qos, force, **kwargs):
        self._ovm3_callback = kwargs['callback']
        kwargs['callback'] = self._mkfs_callback
        if access_grp_names == None:
            access_grp_names = []
        self._ovm3_mkfs = {
            'ss': ss_record.copy(),
            'access_group_names': access_grp_names[:],
            'kwargs': kwargs,
        }
        return self.create_filesystem(self._ovm3_mkfs['ss'], backing_device,
                                      name, size, qos, force, **kwargs)

    def createFileSystem(self, ss_record, name, backing_device, size=0,
                         access_grp_names=None, qos=None, force=False,
                         **kwargs):
        if kwargs.has_key('callback'):
            self._multi_op_start()
            return self._mkfs_async(ss_record, name, backing_device, size,
                                    access_grp_names, qos, force, **kwargs)
        self._multi_op_start()
        fs = self.create_filesystem(ss_record, backing_device, name, size,
                                    qos, force)
        if access_grp_names == None:
            access_grp_names = []
        for ag_name in access_grp_names:
            self.add_elements_to_access_group(ss_record, ag_name, fs)
        fs = self.get_element_info(ss_record, fs)
        self._multi_op_end()
        return fs

    def destroyFileSystem(self, ss_record, fs_record, **kwargs):
        return self.destroy_filesystem(ss_record, fs_record, **kwargs)

    def resizeFileSystem(self, ss_record, fs_record, new_size, **kwargs):
        return self.resize_filesystem(ss_record, fs_record, new_size, **kwargs)

    def listFileSystems(self, ss_record, include_unpresented=False,
                        **kwargs):
        return self.list_filesystems(ss_record, **kwargs)

    def listMountPoints(self, ss_record, fs_record=None, **kwargs):
        return self.list_mountpoints(ss_record, fs_record, **kwargs)


    def mount(self, ss_record, fs_record, mount_point, share_path=None,
              create_mount_point=False, mount_options=[], **kwargs):
        return SCCaller.mount(self, ss_record, fs_record, mount_point,
                              share_path, create_mount_point, mount_options,
                              **kwargs)

    def unmount(self, ss_record, fs_record, mount_point,
                destroy_mount_point=False, **kwargs):
        return SCCaller.unmount(self, ss_record, fs_record, mount_point,
                                destroy_mount_point, **kwargs)

    def getFileSystemBackingDevices(self, ss_record, fs_record, **kwargs):
        return self.get_backing_devices(ss_record, **kwargs)
