#!/usr/bin/python
"""
    Copyright (C) 2010 Oracle.  All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
"""

import sys
from errno import *
import fnmatch
import asyncore
import logging
import time
try:
    import json
except ImportError:
    import simplejson as json

try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
try:
   import cPickle as pickle
except:
   import pickle

from sccaller.chat import StdinoutChat

# Is there a better way to set the import path?
sys.path.append('/opt/storage-connect')
from PluginManager import PluginManager, PluginManagerError
from OSCPlugin import *


__all__ = ['SCCallerWorker']

class SCCallerWorkerHandler(logging.Handler):
    def __init__(self, worker, level=0):
        self._worker = worker
        logging.Handler.__init__(self, level)

    def emit(self, record):
        try:
            message = self.format(record)
            self._worker.push_log(record.levelno, message)
        except Exception, e:
            self.handleError(record)


class SCCallerWorkerError(RuntimeError):
    def __init__(self, errno, message):
        self.errno = errno
        RuntimeError.__init__(self, message)


class SCCallerWorker(StdinoutChat):
    def __init__(self):
        StdinoutChat.__init__(self)
        self.logger = None
        self._handler = None
        self._setup_logger()
        self.plugin_name = None
        self.plugin = None
        self.ss_record = None
        self.manager = PluginManager()
        self.set_terminator('\r\n\r\n')
        self.ibuffer = []
        self.packet_string = ''
        self.packet = None
        self.core_operations = {
            'set-logger': self.op_set_logger,
            'list-plugins': self.op_list_plugins,
            'set-plugin': self.op_set_plugin,
        }
        self.plugin_operations = {
            'get-plugin-attributes': self.op_get_plugin_attributes,
            'validate-server': self.op_validate_server,
            'set-server': self.op_set_server,
        }
        self.server_operations = {
            'get-server-info': self.op_get_server_info,
            'get-server-status': self.op_get_server_status,
            'get-server-attributes': self.op_get_server_attributes,
            'get-storage-names': self.op_get_storage_names,
            'list-luns': self.op_list_luns,
            'list-clones': self.op_list_clones,
            'list-snapshots': self.op_list_snapshots,
            'get-element-status': self.op_get_element_status,
            'get-element-info': self.op_get_element_info,
            'get-element-attributes': self.op_get_element_attributes,
            'list-access-groups': self.op_list_access_groups,
            'create-access-group': self.op_create_access_group,
            'remove-access-group': self.op_remove_access_group,
            'rename-access-group': self.op_rename_access_group,
            'add-hosts-to-access-group': self.op_add_hosts_to_access_group,
            'remove-hosts-from-access-group':
                self.op_remove_hosts_from_access_group,
            'add-elements-to-access-group':
                self.op_add_elements_to_access_group,
            'remove-elements-from-access-group':
                self.op_remove_elements_from_access_group,
            'iscsi-sendtargets': self.op_iscsi_sendtargets,
            'iscsi-start': self.op_iscsi_start,
            'iscsi-stop': self.op_iscsi_stop,
            'lun-scan': self.op_lun_scan,
            'element-online': self.op_element_online,
            'element-offline': self.op_element_offline,
            'get-qos-list': self.op_get_qos_list,
            'set-element-qos': self.op_set_element_qos,
            'create-snapshot': self.op_create_snapshot,
            'create-multi-element_snapshot':
                self.op_create_multi_element_snapshot,
            'create-clone': self.op_create_clone,
            'remove-element': self.op_remove_element,
            'create-lun': self.op_create_lun,
            'resize-element': self.op_resize_element,
            'split-element': self.op_split_element,
            'restore-snapshot': self.op_restore_snapshot,
            'map-devices': self.op_map_devices,
            'create-filesystem': self.op_create_filesystem,
            'destroy-filesystem': self.op_destroy_filesystem,
            'list-filesystems': self.op_list_filesystems,
            'list-mountpoints': self.op_list_mountpoints,
            'create-file': self.op_create_file,
            'destroy-file': self.op_destroy_file,
            'resize-file': self.op_resize_file,
            'clone-file': self.op_clone_file,
            'list-files': self.op_list_files,
            'mount': self.op_mount,
            'unmount': self.op_unmount,
            'list-backing-devices': self.op_list_backing_devices,
            'teardown-device': self.op_teardown_device,
            'resize-device': self.op_resize_device,
            'get-progress': self.op_get_progress,
        }

    def _setup_logger(self, level=0, name=None):
        if self.logger:
            self.logger.removeHandler(self._handler)
            self._handler = None
            self.logger = None
        if not name:
            name = 'SCCallerWorker'
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)
        if level:
            self._handler = SCCallerWorkerHandler(self, level)
            self._handler.setFormatter(logging.Formatter('%(message)s'))
            self.logger.addHandler(self._handler)
        else:
            self.logger.addHandler(NullHandler())
        IPlugin.logger = self.logger

    def log(self, message):
        pass

    def log_info(self, message, type='info'):
        pass

    def collect_incoming_data(self, data):
        self.ibuffer.append(data)

    def found_terminator(self):
        self.packet_string = ''.join(self.ibuffer)
        self.ibuffer = []
        self.packet = json.loads(self.packet_string)
        self.handle_packet()

    def _error_to_errno(self, e):
        '''We don't pass python errors across the communication channel.
           Instead, we pass (errno, message) tuples.'''

        if e.__class__.__name__ == 'SCCallerWorkerError':
            return (e.errno, str(e))
        elif e.__class__.__module__ == 'OSCPlugin':
            return (ENODATA, "%s:%s" % (str(e.__class__.__name__),str(e)))
        elif e.__class__.__name__ == 'ValueError':
            return (EINVAL, str(e))
        elif e.__class__.__name__ == 'TypeError':
            return (ENXIO, str(e))
        elif e.__class__.__name__ == 'NotImplementedError':
            return (EOPNOTSUPP, str(e))
        elif e.__class__.__name__ == 'EnvironmentError':
            return (e.errno, str(e))
        else:
            return (ESRCH, "Unknown error: %s" % str(e))

    def push_result(self, result, e):
        call_error = None
        if e:
            call_error = self._error_to_errno(e)
        result_packet = {
            'magic': 'SCCALLER',
            'major': 1,
            'result': result,
            'error': call_error,
        }
        try:
            self.push('%s\r\n\r\n' % json.dumps(result_packet,
                                                separators=(',',':')))
        except Exception, e:
            sys.stderr.write('Push error "%s"\n' % e)

    def push_cache(self, cache):
        cache_packet = {
            'magic': 'SCCALLER',
            'major': 1,
            'cache': cache,
        }
        try:
            self.push('%s\r\n\r\n' % json.dumps(cache_packet,
                                                separators=(',',':')))
        except Exception, e:
            sys.stderr.write('Push error "%s"\n' % e)

    def push_log(self, level, message):
        log_packet = {
            'magic': 'SCCALLER',
            'major': 1,
            'log_message': message,
            'log_level': level,
        }
        try:
            self.push('%s\r\n\r\n' % json.dumps(log_packet,
                                                separators=(',',':')))
        except Exception, e:
            pass

    def handle_packet(self):
        try:
            self.logger.log(logging.DEBUG / 2,
                            'worker: Got packet "%s"\n' % self.packet_string)
            if self.packet['magic'] != 'SCCALLER':
                raise ValueError('Invalid packet: %s' % self.packet_string)
            if self.packet['major'] != 1:
                raise ValueError('Unknown packet version: %d' %
                                 self.packet['major'])
            op = self.packet['operation']
            args = self.packet['arguments']
            if self.ss_record and self.ss_record.has_key('uuid') and self.ss_record.has_key('cache'):
                if self.ss_record['cache']:
                    cache = pickle.loads(self.ss_record['cache'])
                else:
                    cache = ''
                self.plugin.cache.write(self.ss_record['uuid'], cache)

            result = self.handle_operation(op, args)
            
            if self.ss_record and self.ss_record.has_key('uuid'):
                cache = self.plugin.cache.read(self.ss_record['uuid'])
                if cache:
                    if cache == '':
                        send_cache = {self.ss_record['uuid']: ''} 
                    else:
                        send_cache = {self.ss_record['uuid']:pickle.dumps(cache, 0)} 

                    if op != 'get-progress':
                        self.push_cache(send_cache)
                    else:
                        if type(result) == dict and \
                                result.has_key('async_progress') and \
                                result['async_progress'] == None:
                            self.push_cache(send_cache)
            
            self.push_result(result, None)

        except Exception, e:
            self.logger.debug('worker: Got exception %s: %s' % (e.__class__, e))
            self.logger.debug('Traceback (most recent call last):')
            import traceback
            tb_list = traceback.format_list(traceback.extract_tb(sys.exc_info()[2]))
            for entry in tb_list:
                self.logger.debug("%s" % entry.strip())
            self.push_result(None, e)
        self.packet_string = ''
        self.packet = None

    def handle_operation(self, op, args):
        handler = self.core_operations.get(op, None)
        if not handler:
            handler = self.plugin_operations.get(op, None)
            if handler and not self.plugin:
                raise TypeError('Operation "%s" requires a plugin' % op)
        if not handler:
            handler = self.server_operations.get(op, None)
            if handler:
                if not self.plugin:
                    raise TypeError('Operation "%s" requires a plugin' % op)
                if not self.ss_record:
                    raise TypeError('Operation "%s" requires a server' % op)
        if not handler:
            raise NotImplementedError('No such operation: "%s"' % op)
        return handler(*args)

    def op_list_plugins(self):
        return self.manager.discoverPlugins()

    def op_set_logger(self, level, name):
        self._setup_logger(level, name)

    def op_set_plugin(self, plugin_name):
        if self.plugin:
            raise ValueError('This runner has already loaded the %s plugin' %
                             self.plugin_name)
        if not plugin_name:
            raise TypeError('No plugin name specified')
        try:
            self.plugin = self.manager.getPlugin(plugin_name)
        except PluginManagerError, e:
            raise ValueError(str(e))
        self.plugin_name = plugin_name

    def _info_get(self, cache, info_field):
        info = cache['info']
        return info[info_field]

    def _build_generic_info_map(self, cache):
        attribute_map = {
            'name': lambda : self._info_get(cache, 'plugin_name'),
            'type': lambda : self._info_get(cache, 'plugin_type'),
            'vendor': lambda : self._info_get(cache, 'vendor_name'),
            'description': lambda : self._info_get(cache, 'plugin_desc'),
            'api-version': lambda : self._info_get(cache, 'required_api_vers'),
            'help.extra-info.server':
                lambda : self._info_get(cache, 'ss_extra_info_help'),
        }
        return attribute_map

    def _build_storage_info_map(self, cache):
        attribute_map = {
            'storage.types': lambda : self._info_get(cache, 'storage_types'),
            'storage.api-version':
                lambda : self._info_get(cache, 'sa_api_version'),
            'help.extra-info.element':
                lambda : self._info_get(cache, 'se_extra_info_help'),
        }
        return attribute_map

    def _ability_get(self, cache, ability):
        """Pull a abilty out of the cache, loading the abilities
           if necessary."""
        if not cache.has_key('abilities'):
            cache['abilities'] = self.plugin.getCapabilities(self.ss_record)
        abilities = cache['abilities']
        return abilities[ability]

    def _ability_check(self, cache, ability, test_func):
        """Check a boolean ability with test_func()."""
        ability_value = self._ability_get(cache, ability)
        return test_func(ability_value)

    def _ability_check_chain(self, cache, *args):
        """Check a chain of boolean abilities.  The latter abilities may
           not exist if an earlier ability in the list is False."""
        checks = [(args[i], args[i+1]) for i in range(0, len(args), 2)]
        for (ability, test_func) in checks:
            if not self._ability_check(cache, ability, test_func):
                return False
        return True

    def _ability_get_chain(self, cache, ability, default, *args):
        """Get an ability out of the cache, but only if all the boolean
           abilities in the chain are True.  Return default if a
           dependent ability is False."""
        if self._ability_check_chain(cache, *args):
            return self._ability_get(cache, ability)
        else:
            return default

    def _ability_get_backing_type(self, cache):
        val = self._ability_get(cache, 'backing_device_type')
        if (val == BACKING_DEVICE_TYPES.DEVICE_SINGLE) or \
           (val == BACKING_DEVICE_TYPES.DEVICE_MULTI):
            return 'device'
        elif (val == BACKING_DEVICE_TYPES.PLUGIN_SINGLE) or \
             (val == BACKING_DEVICE_TYPES.PLUGIN_MULTI):
            return 'plugin-specified'
        elif (val == BACKING_DEVICE_TYPES.UNSUPPORTED):
            return 'unsupported'
        else:
            return None

    def _ability_get_backing_multi(self, cache):
        val = self._ability_get(cache, 'backing_device_type')
        if (val == BACKING_DEVICE_TYPES.DEVICE_MULTI) or \
           (val == BACKING_DEVICE_TYPES.PLUGIN_MULTI):
            return True
        else:
            return False

    def _build_filesystem_info_map(self, cache):
        attribute_map = {
            'filesystem.name': lambda : self._info_get(cache, 'filesys_name'),
            'filesystem.type': lambda : self._info_get(cache, 'filesys_type'),
            'filesystem.backing-device.type':
                lambda : self._ability_get_backing_type(cache),
            'filesystem.backing-device.multi':
                lambda : self._ability_get_backing_multi(cache),
            'filesystem.api-version':
                lambda : self._info_get(cache, 'fs_api_version'),
            'help.extra-info.filesystem':
                lambda : self._info_get(cache, 'fs_extra_info_help'),
            'help.extra-info.file':
                lambda : self._info_get(cache, 'file_extra_info_help'),
        }
        return attribute_map

    def _build_ability_map(self, cache):
        online_p = lambda x: x == ABILITY_TYPES.ONLINE
        online_or_offline_p = lambda x: x != ABILITY_TYPES.NO and \
                                        x != ABILITY_TYPES.UNSUPPORTED
        yes_p = lambda x: x == ABILITY_TYPES.YES
        not_yes_p = lambda x: x != ABILITY_TYPES.YES

        attribute_map = {
            'capability.storage-name-required':
                lambda : self._ability_check(cache, 'require_storage_name',
                                             yes_p),
            'capability.snapshot':
                lambda : self._ability_check(cache, 'snapshot',
                                             online_or_offline_p),
            'capability.snapshot.online':
                lambda : self._ability_check(cache, 'snapshot', online_p),
            'capability.snapshot.asynchronous':
                lambda : self._ability_check_chain(cache, 'snapshot',
                                                   online_or_offline_p,
                                                   'snap_is_sync',
                                                   not_yes_p),
            'capability.snapshot.custom-name':
                lambda : self._ability_check_chain(cache, 'snapshot',
                                                   online_or_offline_p,
                                                   'custom_snap_name',
                                                   yes_p),
            'capability.clone':
                lambda : self._ability_check(cache, 'clone',
                                             online_or_offline_p),
            'capability.clone.online':
                lambda : self._ability_check(cache, 'clone', online_p),
            'capability.clone.asynchronous':
                lambda : self._ability_check_chain(cache, 'clone',
                                                   online_or_offline_p,
                                                   'clone_is_sync',
                                                   not_yes_p),
            'capability.clone.custom-name':
                lambda : self._ability_check_chain(cache, 'clone',
                                                   online_or_offline_p,
                                                   'custom_clone_name',
                                                   yes_p),
            'capability.resize':
                lambda : self._ability_check(cache, 'resize',
                                             online_or_offline_p),
            'capability.resize.online':
                lambda : self._ability_check(cache, 'resize', online_p),
            'capability.resize.asynchronous':
                lambda : self._ability_check_chain(cache, 'resize',
                                                   online_or_offline_p,
                                                   'resize_is_sync',
                                                   not_yes_p),
            'capability.splitclone':
                lambda : self._ability_check(cache, 'splitclone',
                                             online_or_offline_p),
            'capability.splitclone.online':
                lambda : self._ability_check(cache, 'splitclone', online_p),
            'capability.splitclone.asynchronous':
                lambda : self._ability_check_chain(cache, 'splitclone',
                                                   online_or_offline_p,
                                                   'splitclone_is_sync',
                                                   not_yes_p),
            'capability.splitclone.open':
                lambda : self._ability_check_chain(cache, 'splitclone',
                                                   online_or_offline_p,
                                                   'splitclone_while_open',
                                                   yes_p),
            'capability.snapclone':
                lambda : self._ability_check(cache, 'snapclone',
                                             online_or_offline_p),
            'capability.snapclone.online':
                lambda : self._ability_check(cache, 'snapclone', online_p),
            'capability.snapclone.asynchronous':
                lambda : self._ability_check_chain(cache, 'snapclone',
                                                   online_or_offline_p,
                                                   'snapclone_is_sync',
                                                   not_yes_p),
            'capability.access-control':
                lambda : self._ability_check(cache, 'access_control', yes_p),
            'capability.access-control.max-entries':
                lambda : int(self._ability_get_chain(cache,
                                                     'max_access_entries',
                                                     0, 'access_control',
                                                     yes_p)),
        }
        return attribute_map

    def _map_attributes(self, object_name, attribute_map, attrs):
        """We only want to query attributes in the map if they are
           requested by the caller.  The keys of the map are all the
           attributes the caller can possibly query.  Once we compute
           the attributes actually being requested, the values of the map
           are thunks that dynamically retrieve the attribute value."""
        attribute_names = attribute_map.keys()
        attr_hash = {}
        for pattern in attrs:
            matched = fnmatch.filter(attribute_names, pattern)
            if not len(matched):
                raise ValueError('Plug-in "%s" does not understand the %s attribute "%s"' %
                                 (self.plugin_name, object_name, pattern))
            for name in matched:
                attr_hash[name] = 1
        results = {}
        for name in attr_hash.iterkeys():
            results[name] = attribute_map[name]()
        return results

    def op_get_plugin_attributes(self, *attrs):
        info = self.plugin.getPluginInfo()
        cache = {
            'info': info,
            'abilities': info['plugin_ability'],
        }
        attribute_map = self._build_generic_info_map(cache)
        if info['plugin_type'] == 'ifs':
            attribute_map.update(self._build_filesystem_info_map(cache))
        elif info['plugin_type'] == 'isa':
            attribute_map.update(self._build_storage_info_map(cache))
        attribute_map.update(self._build_ability_map(cache))
        return self._map_attributes('plug-in', attribute_map, attrs)

    def op_get_server_attributes(self, *attrs):
        info = self.plugin.getPluginInfo()
        cache = {
            'info': info,
        }
        attribute_map = self._build_generic_info_map(cache)
        attribute_map.update(self._build_ability_map(cache))
        if info['plugin_type'] == 'ifs':
            attribute_map.update(self._build_filesystem_info_map(cache))
            if attribute_map['capability.snapshot']():
                attribute_map.update({
                    'capability.snapshot.max':
                        lambda : self.plugin.getFileSystemSnapLimits(self.ss_record),
                })
            if attribute_map['capability.clone']():
                attribute_map.update({
                    'capability.clone.max':
                        lambda : self.plugin.getFileSystemCloneLimits(self.ss_record),
                })
        elif info['plugin_type'] == 'isa':
            attribute_map.update(self._build_storage_info_map(cache))
            if attribute_map['capability.snapshot']():
                attribute_map.update({
                    'capability.snapshot.max':
                        lambda : self.plugin.getSnapLimits(self.ss_record),
                })
            if attribute_map['capability.clone']():
                attribute_map.update({
                    'capability.clone.max':
                        lambda : self.plugin.getCloneLimits(self.ss_record),
                })
        return self._map_attributes('server', attribute_map, attrs)

    def op_validate_server(self, server):
        self.plugin.validate(server)

    def _fill_info(self, ss_record, element=None):

        if hasattr(self.plugin, 'getStorageServerInfo'):
            if element:
                return self.plugin.getFileSystemInfo(ss_record, element)
            else:
                return self.plugin.getStorageServerInfo(ss_record)
        else:
            return self.plugin.getInfo(ss_record, element)

    def op_set_server(self, ss_record):
        self.ss_record = self._fill_info(ss_record)

    def op_get_server_info(self):
        return self._fill_info(self.ss_record)

    def op_get_server_status(self):
        return self.plugin.getStatus(self.ss_record)

    def op_get_storage_names(self):
        return self.plugin.getStorageNames(self.ss_record)

    def op_list_luns(self):
        return self.plugin.list(self.ss_record, [IPlugin.LUNType])

    def op_list_clones(self, *parents):
        clones = []
        if len(parents):
            for parent in parents:
                clones.extend(self.plugin.getCurrentClones(self.ss_record,
                                                           parent))
        else:
            clones = self.plugin.getCurrentClones(self.ss_record)
        return clones

    def op_list_snapshots(self, *parents):
        snapshots = []
        if len(parents):
            for parent in parents:
                snapshots.extend(self.plugin.getCurrentSnaps(self.ss_record,
                                                             parent))
        else:
            snapshots = self.plugin.getCurrentSnaps(self.ss_record)
        return snapshots

    def op_get_element_info(self, element):
        return self._fill_info(self.ss_record, element)

    def op_get_element_status(self, element):
        if hasattr(self.plugin, 'getStorageServerInfo'):
            return self.plugin.getStatus(self.ss_record, element)
        else:
            record = self._fill_info(self.ss_record, element)
            return SE_STATES_TEXT[record['state']]

    def op_get_element_attributes(self, element, *attrs):
        attribute_map = {
            'capability.clone':
                lambda : self.plugin.isCloneable(self.ss_record, element),
            'capability.clone.max':
                lambda : self.plugin.getCloneLimits(self.ss_record, element),
            'capability.snapshot':
                lambda : self.plugin.isSnapable(self.ss_record, element),
            'capability.snapshot.max':
                lambda : self.plugin.getSnapLimits(self.ss_record, element),
            'capability.restore':
                lambda : self.plugin.isRestorable(self.ss_record, element),
            'capability.split':
                lambda : self.plugin.isSplittable(self.ss_record, element),
        }
        return self._map_attributes('element', attribute_map, attrs)

    def op_list_access_groups(self, *elements):
        if len(elements):
            ag_list = {}
            for element in elements:
                ags = self.plugin.getAccessGroups(self.ss_record, element)
                for ag in ags:
                    ag_list[ag['grp_name']] = ag
            ags = ag_list.values()
        else:
            ags = self.plugin.getAccessGroups(self.ss_record)
        return ags

    def op_create_access_group(self, name, *entries):
        ag = {
            'grp_name': name,
            'grp_entries': entries,
        }
        self.plugin.createAccessGroups(self.ss_record, [ag])

    def op_remove_access_group(self, name):
        ag = {
            'grp_name': name,
        }
        self.plugin.removeAccessGroups(self.ss_record, [ag])

    def op_rename_access_group(self, name, new_name):
        self.plugin.renameAccessGroup(self.ss_record, name, new_name)

    def op_add_hosts_to_access_group(self, name, *entries):
        self.plugin.addToAccessGroup(self.ss_record, name, entries)

    def op_remove_hosts_from_access_group(self, name, *entries):
        self.plugin.removeFromAccessGroup(self.ss_record, name, entries)

    def op_add_elements_to_access_group(self, name, *elements):
        ss_record = self.ss_record.copy()
        for se_record in elements:
            self.plugin.startPresent(ss_record, se_record, [name])

    def op_remove_elements_from_access_group(self, name, *elements):
        ss_record = self.ss_record.copy()
        for se_record in elements:
            self.plugin.stopPresent(ss_record, se_record, [name])

    def op_iscsi_sendtargets(self):
        ss_record = self.plugin.discover(self.ss_record)
        return ss_record['storage_id']

    def op_iscsi_start(self, *targets):
        ss_record = self.ss_record.copy()
        ss_record['storage_id'] = targets
        self.plugin.start(ss_record)

    def op_iscsi_stop(self, *targets):
        ss_record = self.ss_record.copy()
        ss_record['storage_id'] = targets
        self.plugin.stop(ss_record)

    def op_lun_scan(self, *targets):
        ss_record = self.ss_record.copy()
        ss_record['storage_id'] = targets
        self.plugin.refresh(ss_record)

    def op_map_devices(self, element):
        ss_record = self.ss_record.copy()
        records = self.plugin.updateSERecords(ss_record, [element], False)
        return records[0]['path']

    def op_element_online(self, element):
        ss_record = self.ss_record.copy()
        self.plugin.online(ss_record, element)

    def op_element_offline(self, element):
        ss_record = self.ss_record.copy()
        self.plugin.offline(ss_record, element)

    def op_get_qos_list(self, element):
        ss_record = self.ss_record.copy()
        if element:
            return self.plugin.getQoSList(ss_record, element)
        else:
            return self.plugin.getQoSList(ss_record)

    def op_set_element_qos(self, element, qos_value):
        ss_record = self.ss_record.copy()
        self.plugin.setQoS(ss_record, element, qos_value)

    def op_create_snapshot(self, element, snapshot_name):
        ss_record = self.ss_record.copy()
        snap_se_record = None
        if snapshot_name:
            # id is a hack for older plugins
            snap_se_record = dict(name=snapshot_name, id=snapshot_name)
        op_record = self.plugin.createSnap(ss_record, element,
                                           snap_se_record)
        caps = self.plugin.getCapabilities(self.ss_record)
        #if caps.get('snap_is_sync', ABILITY_TYPES.YES) != ABILITY_TYPES.YES:
            #op_record = self._watch_async_progress(op_record)
        return op_record

    def op_create_multi_element_snapshot(self, elements, snapshot_name):
        ss_record = self.ss_record.copy()
        snap_se_record = None
        if snapshot_name:
            # id is a hack for older plugins
            snap_se_record = dict(name=snapshot_name, id=snapshot_name)
        return self.plugin.createMultiSnap(ss_record, elements,
                                           snap_se_record)

    def op_create_clone(self, element, clone_name, qos_value,
                        snapshot_element, plugin_options):
        ss_record = self.ss_record.copy()
        if clone_name or plugin_options:
            clone_se_record = {}
            if clone_name:
                # id is a hack for older plugins
                clone_se_record['name'] = clone_name
                clone_se_record['id'] = clone_name
            if plugin_options:
                clone_se_record['extra_info'] = plugin_options
        else:
            clone_se_record = None
        if snapshot_element:
            sync_cap = 'snapclone_is_sync'
            op_record = self.plugin.cloneFromSnap(ss_record, element,
                                                  snapshot_element,
                                                  clone_se_record,
                                                  qos_value)
        else:
            sync_cap = 'clone_is_sync'
            op_record = self.plugin.clone(ss_record, element,
                                          clone_se_record, qos_value)
        caps = self.plugin.getCapabilities(self.ss_record)
        #if caps.get(sync_cap, ABILITY_TYPES.YES) != ABILITY_TYPES.YES:
            #op_record = self._watch_async_progress(op_record)
        return op_record

    def op_remove_element(self, element):
        ss_record = self.ss_record.copy()
        if element['se_type'] == IPlugin.SnapType:
            return self.plugin.snapRemove(ss_record, element)
        else:
            return self.plugin.remove(ss_record, element)

    def op_create_lun(self, volume_group, name, size, thin_provisioned,
                      qos_value, plugin_options):
        ss_record = self.ss_record.copy()
        se_record = dict(name=name, size=size, vol_group_name=volume_group)
        if plugin_options:
            se_record['extra_info'] = plugin_options
        return self.plugin.create(ss_record, se_record, thin_provisioned,
                                  qos_value)

    def op_resize_element(self, element, new_size):
        ss_record = self.ss_record.copy()
        if hasattr(self.plugin, 'getStorageServerInfo'):
            op_record = self.plugin.resizeFileSystem(ss_record, element,
                                                     new_size)
        else:
            op_record = self.plugin.resize(ss_record, element, new_size)
        caps = self.plugin.getCapabilities(self.ss_record)
        #if caps.get('resize_is_sync', ABILITY_TYPES.YES) != ABILITY_TYPES.YES:
            #self._watch_async_progress(op_record)

    def _watch_async_progress(self, op_record):
        ss_record = self.ss_record.copy()
        while op_record.get('async_progress', None) != None:
            time.sleep(1)
            op_record = self.plugin.getAsyncProgress(ss_record, op_record)
        return op_record

    def op_split_element(self, element):
        ss_record = self.ss_record.copy()
        op_record = self.plugin.splitClone(ss_record, element)
        caps = self.plugin.getCapabilities(self.ss_record)
        #if caps.get('splitclone_is_sync', ABILITY_TYPES.YES) != ABILITY_TYPES.YES:
            #self._watch_async_progress(op_record)

    def op_restore_snapshot(self, snapshot, element):
        ss_record = self.ss_record.copy()
        self.plugin.snapRestore(ss_record, snapshot, element)

    def op_create_filesystem(self, backing_device, name, size, qos_value,
                             force, plugin_options):
        ss_record = self.ss_record.copy()
        return self.plugin.createFileSystem(ss_record, name, backing_device,
                                            size, None, qos_value, force)

    def op_destroy_filesystem(self, fs):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        self.plugin.destroyFileSystem(ss_record, fs_record)

    def op_list_filesystems(self):
        return self.plugin.listFileSystems(self.ss_record, True)

    def op_list_mountpoints(self, fs):
        return self.plugin.listMountPoints(self.ss_record, fs)

    def op_create_file(self, fs, file_name, file_type, size,
                             sparse, force):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        return self.plugin.create(ss_record, fs_record, file_name,
                                  file_type, size, sparse, force)

    def op_resize_file(self, fs, file, size, sparse):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        file_record = file.copy()
        return self.plugin.resize(ss_record, fs_record, file_record,
                                  size, sparse)

    def op_destroy_file(self, fs, file):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        file_record = file.copy()
        return self.plugin.destroy(ss_record, fs_record, file_record)

    def op_clone_file(self, fs, file, clone_name):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        file_record = file.copy()
        return self.plugin.clone(ss_record, fs_record, file_record,
                                   clone_name)

    def op_list_files(self, fs, file, recursive):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        file_record = file.copy()
        return self.plugin.list(ss_record, fs_record, file_record, recursive)
    
    def op_mount(self, fs, mountpoint, share_path, create_mountpoint,
                 mount_options):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        return self.plugin.mount(ss_record, fs_record, mountpoint,
                                 share_path, create_mountpoint,
                                 mount_options)

    def op_unmount(self, fs, mountpoint, destroy_mountpoint):
        ss_record = self.ss_record.copy()
        fs_record = fs.copy()
        return self.plugin.unmount(ss_record, fs_record, mountpoint,
                                   destroy_mountpoint)

    def op_list_backing_devices(self):
        return self.plugin.getBackingDevices(self.ss_record, {})

    def op_teardown_device(self, se):
        ss_record = self.ss_record.copy()
        se_record = se.copy()
        return self.plugin.deviceTeardown(ss_record, se_record)

    def op_resize_device(self, se):
        ss_record = self.ss_record.copy()
        se_record = se.copy()
        return self.plugin.deviceResize(ss_record, se_record)

    def op_get_progress(self, result_record):
        ss_record = self.ss_record.copy()
        return self.plugin.getAsyncProgress(ss_record, result_record)
         
if __name__ == '__main__':
    def main(args):
        chat = SCCallerWorker()
        # SCCaller dup()s the descriptors.  Close the originals so that
        # no other prints get to the communication channel.
        sys.stdin.close()
        sys.stdout.close()
        asyncore.loop()
        sys.exit(0)

    sys.exit(main(sys.argv))
